{{
/**
 *******************************************************************************************************
 * @file jvmdoc.hh
 *
 * @note Jazzed_JVM_Demo_008.spin
 *      Added public front end API.
 *      Added debug menu for API testing and JVM debug.
 *      Issues: menu command i causes r & c work only once per boot.
 *      WorkAround: avoid using i command :).
 *
 * @note Jazzed_JVM_Demo_006.spin.
 *      Added handling for InvokeVirtual.
 *      Added Peter's native method code.
 *      Hello World program works :).
 *
 * @note Jazzed_JVM_Demo_004.spin.
 *      Added support for LDC and siblings.
 *      Added support for GETFIELD/PUTFIELD.
 *      Added support for INSTANCEOF.
 *      Added support for NEW, NEWARRAY, and ANEWARRAY.
 *      Added code to InvokeProgram to get us further along.
 *
 * @note Jazzed_JVM_Demo_002.spin.
 *      Merged debug support with AiChip_JVM_Demo_002.spin.
 *
}}
PRI Version
  return String("Jazzed_JVM_Demo_008")

CON
  

DAT ' jazzDebug , debugMsDelay < 0 means stop for debug
  jazzDebug             long 0
  callDebug             long 0
  natDebug              long 0
  debugging             long 0 ' Set true if debugging the JVM
  debugMsDelay          long 0
  gAbort                long 0 ' allows user to abort JVM run if set
                               
CON
{  
As supplied the jem.out is ...

'        import stamp.*;
'        public class testJVM
'        {
'          static int iStatic;
'          static int incr( int a )
'          {
'            return a + 1;
'          }
'          static void main()
'          {
'            iStatic = 0;
'            while (true)
'            {
'              iStatic = java.lang.Math.min( incr(iStatic), 10);
'              stamp.core.CPU.delay(iStatic);
'            }
'          }
'        }

Progress so far ...

        DONE - Get 'main' running by itself using local variables
        DONE - Get 'main' calling other class native methods
        DONE - Get 'main' calling same class static method
        DONE - Get 'main' calling other class static methods
        DONE - Get 'main' called from program initialisation
        DONE - Create simple heap handler
        DONE - Handle array creation
        DONE - Handle class static variables for 'testJVM'
        DONE - Get main running with class static variables for 'testJVM'

        TODO - Add Native Method handling code
        Added Peter's native library, handler, and parameter passing ... jazzed
        
Not Implemented yet ...

        Jem_INVOKEVIRTUAL       jazzed
        Jem_INVOKEINTERFACE
        Jem_ATHROW       
        Jem_CHECKCAST
        Jem_INSTANCEOF          jazzed
        Jem_WIDE
        
        Jem_GETFIELD            jazzed 
        Jem_PUTFIELD            jazzed 
        Jem_NEW                 jazzed
        
        Jem_ANEWARRAY           jazzed
        
        Jem_LDC                 jazzed
        Jem_LDC_W
        Jem_LDC2_W

        Jem_TABLESWITCH
        Jem_LOOKUPSWITCH

        PopObjectRef            Need for IF_ACMPEQ/NE
      
Not completely implemented ...

        InvokeProgram           Just grabs an array for static fields ...  jazzed
        Jem_NEWARRAY            What does the operand mean ? Always $00 !  jazzed

}
CON
{

PLEASE NOTE : If modifying and re-releasing this code, please change the prefix of the source file from
              AiChip so as to avoid any confusion as to the origin of the source and update all internal
              references as appropriate. If you wish to credit the orginal source, anything  along the
              lines of, "Based on original AiChip_Jvm_XXX from AiChip Industries", is acceptable.

              AiChip is a trademark of AiChip Industries.

}
CON

' .-----------------------------------------------------------------------------------------------------.
' |     Board Configuration                                                                             |
' `-----------------------------------------------------------------------------------------------------'

  _CLKMODE            = XTAL1 + PLL16x                  ' 80MHz System Clock Speed
  _XINFREQ            = 5_000_000

OBJ

  'tv                 : "VGA_Text"                      ' TV Text Display used for debugging
  'tv                 : "Jazzed_JVM_MonDebug"           ' TV Text Display used for debugging
  jem                 : "AiChip_JVM_JemFile_002.spin"   ' JEM Bytecode to execute ( jem.bin )
  tv                  : "Jazzed_SIO_Text"               ' Serial IO
  
CON

' .-----------------------------------------------------------------------------------------------------.
' |     Hardware Configuration                                                                          |
' `-----------------------------------------------------------------------------------------------------'

  TV_PIN              = 16 '12                              ' Base TV Pin

CON

' .-----------------------------------------------------------------------------------------------------.
' |     JVM Configuration                                                                               |
' `-----------------------------------------------------------------------------------------------------'

  INT32               = false                           ' Set true if an int is 32-bit, otherwise 16-bit
  LNG64               = false                           ' Set true if a long is 64-bit, otherwise 32-bit
  CHR16               = true                            ' Set true if a char is 16-bit, otherwise 8-bit

  BIGARRAYS           = false                           ' Set true if allowed arrays larger than 64KB
  
  STACK_SIZE          = 1024                            ' Entries allowed on JVM runtime execution stack

  HEAP_BASE           = $7E00                           ' Address of first free byte above top of heap

CON
  
  INTWORDS            = 1 - INT32                       ' Number of words in an int ( INT32=2, INT16=1 )
  INTBYTES            = INTWORDS * 2                    ' Number of bytes in an int ( INT32=4, INT16=2 )

  LNGWORDS            = ( 1 - LNG64 ) * 2               ' Number of words in a long ( LNG64=4, LNG32=2 )
  LNGBYTES            = LNGWORDS * 2                    ' Number of bytes in a long ( LNG64=8, LNG32=4 )

  CHRWORDS            = ( 1 - CHR16 ) >> 1              ' Number of words in a char ( CHR16=1, CHR8=0  )
  CHRBYTES            = 1 - CHR16                       ' Number of bytes in a char ( CHR16=2, CHR8=1  )

  MODE_SPIN_ONLY      = %010                            ' How the JVM opcode processing can be handled
  MODE_SPIN_THEN_PASM = %011
  MODE_PASM_ONLY      = %100
  MODE_PASM_THEN_SPIN = %110

  
  HEAP_HEAD_TYPE      = 0                               ' Heap chunk Type entry offset
  HEAP_HEAD_BLEN      = 1                               ' Heap chunk ByteLen entry offset
  HEAP_HEAD_DATA      = 2                               ' Heap chunk Data entry offset

  HEAP_WORD           = 1
  HEAP_LONG           = 2

  BREAKPOINT_SIZE     = 10

CON 'jvm instruction termination codes

  #-3, jvmCrashed, jvmAborted, jvmBreakPoint, jvmExitOk, jvmPasmModeDisabled, jvmSpinModeDisabled  

  SBUFLEN = 128
  
VAR

' .-----------------------------------------------------------------------------------------------------.
' |     JVM Runtime Data Structures                                                                     |
' `-----------------------------------------------------------------------------------------------------'

  ' The JVM runtime stack is adjusted for the base
  ' size of int(16) or int(32). Only half as many
  ' longs will be used for int(16) as for int(32).
  ' Keep this as long so it works with INT32.
  
  long stack[ STACK_SIZE >> (INT32+1) ]                 ' Stack for JVM runtime execution

  ' The following ( 'pc' to 'sp' ) must all be longs
  ' and must be contiguous. Add new stack frame items
  ' within this list, not before 'pc' nor after 'sp'
  ' or the stack frames will not work properly.
  
  long pc                                               ' Current JEM bytecode Pc ( absilute hub address )
  long bp                                               ' Current base of stack ( firat param / local )
  long mb                                               ' Method base - first executable in method
  long fp                                               ' Current frame pointer
  long ct                                               ' Constant table address
  long et                                               ' Exception table address
  long vp                                               ' Variable base ( static fields )
  long sp                                               ' Top of stack pointer ( last pushed )


  long bytecodeStart                                    ' Absolute Hub address of first bytecode
  long jemFileCountAdjust                               ' How to adjust Jem File counts to fit reality

  long heap                                             ' Ptr to last chunk created on heap
  
  long mode                                             ' Type of JVM opcode processing
  long allowNotImplementedNatives                       ' Set true if unimplemented Native Methods ignored

  long stackAltered                                     ' True if stack has altered flag
  long opcode                                           ' Current opcode

  long gBytecodeLen                                     ' length of bytecode (len field * 256)
  long gBytecodeMain                                    ' bytecode file's address for main
  long gClassId                                         ' working object class id
  long gMethodId                                        ' working method id
  long gObjectClassId                                   ' java.lang.Object class id
  long gConstTable                                      ' constant table pointer

  word gGetOverBreak                                    ' skip break once if set during isBreakPoint
  word gBreakpointHead                                  ' breakpoint array queue head
  word gBreakpointArray[BREAKPOINT_SIZE]                ' breakpoint array

  'long cog
  'long cogstack[600]


PUB Main | last, ofp, omp, sbuf[SBUFLEN+1], printing
{{              
 *******************************************************************************************************
 */

/**
 * Main entry point for standalone jvm testing.
 */
__ void main(); 
}}
  start

  clearAllJvmBreakpoints

  printMenu

  jvmInit
  
  printing := 0

  repeat
    tv.Str(String($d,"JVM> "))
    result := tv.getch
    if result <> $D
      last := result
      if last <> "p" and printing
        printing := 0
    tv.out(last)
    case last
      "b": ' set breakpoint
        tv.Str(String(" jem.out line number $> "))
        result := getBreakAddr
        if result <> 0
          tv.hex(result,4)
          setJvmBreakpoint(result+bytecodeStart)
        result := 0
      "c": ' continue debugging
        repeat
          if tv.key
            quit
          if jvmSpinDebug(1) < 0
            quit
      "C": ' clear all breakpoints
        clearAllJvmBreakpoints
      "d": ' delete a given breakpoint
        tv.Str(String("delete bp $> "))
        result := getBreakAddr
        if result <> 0
          tv.hex(result,4)
          deleteJvmBreakpoint(result+bytecodeStart)
      "f": ' singlestep until frame changes
        ofp := getJvmFP
        repeat
          if jvmSpinDebug(1) < 0
            quit
          if tv.key <> 0
            tv.getch
            quit
        while getJvmFP == ofp
      "h": ' dump heap
        ShowHeap
      "i": ' reinit jvm
        jvmInit
      "l": ' list all breakpoints
        listJvmBreakpoints
      "m": ' singlestep until method changes
        omp := getJvmMP
        repeat
          if jvmSpinDebug(1) < 0
            quit
          if tv.key <> 0
            tv.getch
            quit
        while getJvmMP == omp
      "n": ' singlestep debug
        gGetOverBreak := 1
        jvmSpinDebug(1)
      "o": ' dump an object
        tv.Str(String(" address $> "))
        tv.fgets(@sbuf, SBUFLEN)
        result := tv.strtohex(@sbuf)
        ShowObject(result)      
      "p": ' print memory
        if printing == 0
          tv.Str(String(" address $> "))
          tv.fgets(@sbuf, SBUFLEN)
          result := tv.strtohex(@sbuf)
          dumpBytes(result,$100)
        else      
          dumpBytes(printing,$100)
        printing := result+$100
      "q": ' reboot
        reboot
      "r": ' run to a breakpoint
        pc := gBytecodeMain+bytecodeStart
        tv.out($d)   
        repeat
          if tv.key
            quit
          if jvmSpinDebug(1) < 0
            quit
      "s": ' dump stack
        ShowStack
      "v": ' set callDebug verbosity
        callDebug := 1
      "V": ' set unintelligible verbosity 
        debugging := 1  
      "u": ' unset verbosity
        callDebug := 0 
        debugging := 0
      "x": ' x-amine memory location hex value
        tv.Str(String(" address $> "))
        tv.fgets(@sbuf, SBUFLEN)
        result := tv.strtohex(@sbuf)
        tv.Str(string("address "))
        tv.hex(result, 4)
        tv.Str(string(" = $"))
        tv.hex(word[result], 2)
      other:
        printMenu

PUB dumpBytes(addr,len) | ndx
{{
/**
 * dump bytes in 8 column table
 * @param addr is base address for dump
 * @param len is number of bytes to dump
 */
__ void dumpBytes(addr,len);
}}
  ndx := 0
  'str(string("Dump"))
  repeat while(ndx < len)
    if ndx <> 0 and (ndx // 16) <> 0
      tv.out(".")
    if ndx == 0 or (ndx // 16) == 0
      tv.out($d)
      tv.hex(addr+ndx,4)
      tv.str(string(": "))
    tv.hex(byte[addr+ndx++],2)
      
PRI getBreakAddr | sbuf[SBUFLEN+1]
  tv.fgets(@sbuf, SBUFLEN)
  result := tv.strtohex(@sbuf)
  
  if result < gBytecodeMain
    tv.Str(String("PC < jem.out main address. Try again. "))
    tv.hex(result,4)
    tv.out(" ")
    tv.hex(gBytecodeMain,4)
    result := 0      
  elseif result > gBytecodeLen
    tv.Str(String("PC exceeds jem.out size. Try again. "))
    tv.hex(result,4)
    tv.out(" ")
    tv.hex(gBytecodeLen,4)
    result := 0      

PRI printMenu
  tv.cls 
  tv.out($d)
  tv.str( Version )
  tv.out($d)
  tv.str(string($d,"JVM Human Interface Debug Help:"))
  tv.str(string($d," b: set breakpoint"))
  tv.str(string($d," c: continue debugging"))
  tv.str(string($d," C: clear all breakpoints"))
  tv.str(string($d," d: delete a given breakpoint"))
  tv.str(string($d," f: finish a method"))
  tv.str(string($d," h: dump heap"))
  tv.str(string($d," i: init JVM to reset state"))
  tv.str(string($d," l: list all breakpoints"))
  tv.str(string($d," m: break on new method"))
  tv.str(string($d," n: single step"))
  tv.str(string($d," o: dump an object contents.")) 
  tv.str(string($d," p: print memory range.")) 
  tv.str(string($d," q: quit and reboot"))
  tv.str(string($d," r: run to a breakpoint"))  
  tv.str(string($d," s: dump stack"))
  tv.str(string($d," v: set callDebug verbosity")) 
  tv.str(string($d," V: set more detail verbosity")) 
  tv.str(string($d," u: unset verbosity")) 
  tv.str(string($d," x: examine one memory address.")) 
  tv.str(string($d," Enter repeats last command."))
  tv.str(string($d," Use 'n' to skip breakpoint."))
  tv.out($d)
  
PUB start
{{
/**
 * This is the startup point for module using the JVM
 * A new COG is not initialized for this today.
 */
__ void start();
}}
  'tv.init
  'tv.Start( TV_PIN )

  tv.start(31, 30, 0, 9600)
  'if debugMsDelay < 0
  '  repeat while(tv.key == 0)

  mode := MODE_SPIN_ONLY

  allowNotImplementedNatives := true
  gConstTable := 0
  gBreakpointHead := 0
  heap := 0
  
  'jvmInit
  'JvmSpinLoop

  'cog := cognew(mainrun, @cogstack) + 1
  
PUB stop
{{
/**
 * This is a compatibility stop function.
 * A new COG is not initialized for the module, so nothing is stopped.
 */
__ void stop();
}}
  'if cog
  '   cogstop(cog~ -  1)

'PRI mainrun

PUB jvmInit | ii
{{
/**
 * Run the JVM init procedure. This resets the JVM engine.
 */
__ void jvmInit();
}}

' .-----------------------------------------------------------------------------------------------------.
' |     Initialise JVM                                                                                  |
' `-----------------------------------------------------------------------------------------------------'

  sp := @stack - INTBYTES
  bp := sp
  mb := 0
  fp := 0
  pc := 0
  if heap <> 0
    repeat ii from heap to HEAP_BASE
      byte[ii] := 0  
  heap := HEAP_BASE
  gAbort := false
  InvokeProgram( jem.BytecodeStart )

PRI dumpRegs
  tv.Out( $0D )
  tv.str(string("PC "))
  tv.hex(getJvmPC, 4)
  tv.str(string(" SP "))
  tv.hex(getJvmSP, 4)
  tv.str(string(" SB "))
  tv.hex(getJvmSB, 4)
  tv.str(string(" FP "))
  tv.hex(getJvmFP, 4)
  tv.str(string(" MP "))
  tv.hex(getJvmMP, 4)
  tv.str(string(" HP "))
  tv.hex(getJvmHP, 4)
  tv.str(string(" OP "))
  tv.hex(gClassId, 4)
  'tv.Out($d)
  tv.str(string(" PC "))
  'tv.Hex( pc, 4 )
  'tv.Out( " " )
  tv.Hex( pc - bytecodeStart, 4 )
  tv.Out( ":" )
  tv.hex( getJvmMP-bytecodeStart, 4)
  tv.Out( " " )
  tv.Dec( pc - mb )
  if (pc - mb ) => 0 and ( pc - mb ) < 10
    tv.Out( " " )
  if ( pc - mb ) < 100
    tv.Out( " " )
   
    
PRI JvmSpinDebug(singlestep) | ch
{{
/**
 * Runs the JVM until a breakpoint is hit or if singlestep is non-zero runs one instruction.
 * @returns 0 if function terminates normally.
 */
__ int jvmSpinDebug(singlestep);
}}
' .-----------------------------------------------------------------------------------------------------.
' |     Main JVM Loop                                                                                   |
' `-----------------------------------------------------------------------------------------------------'

  repeat

    if gAbort
      gAbort := false
      return 0
            
    if debugging
      dumpRegs

    if isBreakpoint(pc)
      return jvmBreakPoint

    opcode := byte[ pc++ ]
     
    if debugging
      tv.Hex( opcode, 2 )
      tv.Str( string(" - "))
      tv.Str( jem.OpcodeNamePtr( opcode ) )
      tv.Str( string(" - "))
     
    result := \JvmSpin
    if result <> 0
      gAbort := false
      return result
     
    'if debugging
      'ShowStack

    if singlestep
      return 0


PRI JvmSpinLoop | ch

' .-----------------------------------------------------------------------------------------------------.
' |     Main JVM Loop                                                                                   |
' `-----------------------------------------------------------------------------------------------------'

  repeat

    if gAbort
      gAbort := false
      return 0
  
    if debugging

      ShowStack
      if tv.key == $D
        debugMsDelay := -1
      if debugMsDelay < 0
        ch := tv.getch
        if ch == $20
          debugMsDelay := 0
        'else \deb.menu(ch)
        
      elseif debugMsDelay > 0
        Pause_Ms( debugMsDelay )

      tv.Out( $0D )
      tv.str(string("PC "))
      'tv.Hex( pc, 4 )
      'tv.Out( " " )
      tv.Hex( pc - bytecodeStart, 4 )
      tv.Out( ":" )
      tv.Dec( pc - mb )
      if (pc - mb ) => 0 and ( pc - mb ) < 10
        tv.Out( " " )
      if ( pc - mb ) < 100
        tv.Out( " " )
   
      opcode := byte[ pc++ ]

      tv.Hex( opcode, 2 )
      tv.Str( string(" - "))
      tv.Str( jem.OpcodeNamePtr( opcode ) )
      tv.Str( string(" - "))

      result := \JvmSpin
      if result <> 0
        gAbort := false
        quit
       
    else
  
      opcode := byte[ pc++ ]
      result := \JvmSpin
      if result <> 0
        gAbort := false
        quit
      
    if debugging
      ShowStack

  'Crashed( String( "JvmSpin terminated" ) )

    
PRI JvmSpin | arrayPtr, arrayLen, arrayIndex, value

  if gAbort <> 0
    return jvmAborted
    
  if mode & MODE_PASM_ONLY
    if JvmPasm
       return jvmPasmModeDisabled
    else
      if mode == MODE_PASM_ONLY
        return jvmSpinModeDisabled
        'Crashed( String( "PASM did not handle opcode" ) )

  if mode & MODE_SPIN_ONLY
              
    if opcode < jem#Jem_INSTANCEOF
      case opcode

        ' .---------------------------------------------------------------------------------------------.
        ' |     NOP                                                                                 $00 |
        ' |     Meaning: Do nothing.                                                                    |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_NOP:

        ' .---------------------------------------------------------------------------------------------.
        ' |     ICONST_M1                                                                           $01 |
        ' |     Meaning: Push the integer constant -1.                                                  |
        ' |     Description: This bytecode pushes the integer constant -1 onto the stack.               |
        ' |---------------------------------------------------------------------------------------------|
        ' |     ICONST_0                                                                        $02-$07 |
        ' |     Meaning: Push the integer constant <C>.                                                 |
        ' |     Description: This bytecode pushes an integer constant onto the stack. There             |
        ' |     are six associated bytecodes, one for each of the integers 0-5: iconst_0,               |
        ' |     iconst_1, iconst_2, iconst_3, iconst_4, and iconst_5.                                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ICONST_M1, jem#Jem_ICONST_0, jem#Jem_ICONST_1, jem#Jem_ICONST_2, jem#Jem_ICONST_3:

          Push( opcode - jem#Jem_ICONST_0 )

        jem#Jem_ICONST_4, jem#Jem_ICONST_5:

          Push( opcode - jem#Jem_ICONST_4 )

        ' .---------------------------------------------------------------------------------------------.
        ' |     BIPUSH                                                                              $08 |
        ' |     Meaning: Push a 1-byte signed integer.                                                  |
        ' |     Description: The first byte following the opcode, byte1, is interpreted as a            |
        ' |     signed 8-bit value. This value is expanded to a 32-bit integer and pushed onto          |
        ' |     the operand stack.                                                                      |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_BIPUSH:

          Push( Operand_SignedByte )

        ' .---------------------------------------------------------------------------------------------.
        ' |     SIPUSH                                                                              $09 |
        ' |     Meaning: Push a 2-byte signed integer.                                                  |
        ' |     Description: The first and second byte following the opcode, byte1 and byte2,           |
        ' |     are assembled into a signed 16-bit value. This value is expanded to a 32-bit            |
        ' |     integer and pushed onto the operand stack.                                              |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_SIPUSH:

          Push( Operand_SignedWord )

        ' .---------------------------------------------------------------------------------------------.
        ' |     ILOAD                                                                               $0A |
        ' |     Meaning: Load integer from local variable.                                              |
        ' |     Description: The value of the local variable in the current Java frame is               |
        ' |     pushed onto the operand stack. This value must be of type integer.                      |
        ' |---------------------------------------------------------------------------------------------|
        ' |     ILOAD_0                                                                             $0B |
        ' |     ILOAD_1                                                                             $0C |
        ' |     ILOAD_2                                                                             $0D |
        ' |     ILOAD_3                                                                             $0E |
        ' |     Meaning: Load integer from local variable with specific index.                          |
        ' |     Description: The value of the local variable <l> in the current Java frame is           |
        ' |     pushed onto the operand stack.                                                          |
        ' |     This value must be of type integer. There are four of these bytecodes, one for          |
        ' |     each of the integers 0-3: iload_0, iload_1, iload_2, and iload_3.                       |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ILOAD:

           Push( GetVar( Operand_UnsignedByte ) )

        jem#Jem_ILOAD_0, jem#Jem_ILOAD_1, jem#Jem_ILOAD_2, jem#Jem_ILOAD_3:

           Push( GetVar( opcode - jem#Jem_ILOAD_0 ) )

        ' .---------------------------------------------------------------------------------------------.
        ' |     ISTORE                                                                              $0F |
        ' |     Meaning: Store integer into local variable.                                             |
        ' |     Description: Local variable in the current Java frame is set to a value. This           |
        ' |     value must be an integer.                                                               |
        ' |---------------------------------------------------------------------------------------------|
        ' |     ISTORE_0                                                                            $10 |
        ' |     ISTORE_1                                                                            $11 |
        ' |     ISTORE_2                                                                            $12 |
        ' |     ISTORE_3                                                                            $13 |
        ' |     Meaning: Store integer into local variable with specific index.                         |
        ' |     Description: Local variable <l> in the current Java frame is set to a value.            |
        ' |     This value must be an integer. There are four of these bytecodes, one for each          |
        ' |     of the integers 0-3: istore_0, istore_1, istore_2, and istore_3.                        |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ISTORE:

           PutVar( Operand_UnsignedByte, Pop )

        jem#Jem_ISTORE_0, jem#Jem_ISTORE_1, jem#Jem_ISTORE_2, jem#Jem_ISTORE_3:

           PutVar( opcode - jem#Jem_ISTORE_0, Pop )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IFxx                                                                                $14 |
        ' |     Meaning: Branch if condition                                                            |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IFEQ: BranchIf( Pop == 0 )
        jem#Jem_IFNE: BranchIf( Pop <> 0 )
        jem#Jem_IFLT: BranchIf( Pop <  0 )
        jem#Jem_IFLE: BranchIf( Pop =< 0 )
        jem#Jem_IFGT: BranchIf( Pop >  0 )
        jem#Jem_IFGE: BranchIf( Pop => 0 )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IF_ICMPxx                                                                           $1A |
        ' |     Meaning: Branch if compare                                                              |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IF_ICMPEQ: BranchIf( PopL == PopR )
        jem#Jem_IF_ICMPNE: BranchIf( PopL <> PopR )
        jem#Jem_IF_ICMPLT: BranchIf( PopL <  PopR )
        jem#Jem_IF_ICMPGE: BranchIf( PopL => PopR )
        jem#Jem_IF_ICMPGT: BranchIf( PopL >  PopR )
        jem#Jem_IF_ICMPLE: BranchIf( PopL =< PopR )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IF_ACMPxx                                                                           $20 |
        ' |     Meaning: Branch if object references equal.                                             |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IF_ACMPEQ: BranchIf( PopObjectRef == PopObjectRef )
        jem#Jem_IF_ACMPNE: BranchIf( PopObjectRef <> PopObjectRef )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IRETURN                                                                             $22 |
        ' |     Meaning: Return integer from function.                                                  |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IRETURN: ReturnFromMethodWithValue
        jem#Jem_RETURN:  ReturnFromMethod

        ' .---------------------------------------------------------------------------------------------.
        ' |     INVOKEVIRTUAL                                                                       $24 |
        ' |     Meaning: Invoke method based on runtime type.                                           |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_INVOKEVIRTUAL:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     INVOKESTATIC                                                                        $25 |
        ' |     Meaning: Invoke a class (static) method.                                                |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_INVOKESTATIC:

          InvokeStatic( Operand_BytecodeAddress )

        ' .---------------------------------------------------------------------------------------------.
        ' |     GETSTATIC                                                                           $26 |
        ' |     Meaning: Get 16bit static field from class.                                             |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_GETSTATIC:

          Push( Peek( vp, Operand_UnsignedWord >> jemFileCountAdjust ) )

        ' .---------------------------------------------------------------------------------------------.
        ' |     PUTSTATIC                                                                           $27 |
        ' |     Meaning: Set 16bit static field in class.                                               |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_PUTSTATIC:

          Poke( vp, Operand_UnsignedWord >> jemFileCountAdjust, Pop )

        ' .---------------------------------------------------------------------------------------------.
        ' |     POP                                                                                 $28 |
        ' |     Meaning: Stack manipulation                                                             |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_POP:     Pop
        jem#Jem_POP2:    Pop2
        jem#Jem_DUP:     Dup
        jem#Jem_DUP_X1:  Dup_X1
        jem#Jem_DUP_X2:  Dup_X2
        jem#Jem_DUP2:    Dup2
        jem#Jem_DUP2_X1: Dup2_X1
        jem#Jem_DUP2_X2: Dup2_X2
        jem#Jem_SWAP:    Swap

        ' .---------------------------------------------------------------------------------------------.
        ' |     IADD, ISUB, INEG, IOR, ISHL, ISHR, IUSHR, IAND, IXOR                                $31 |
        ' |     Meaning: Integer maths                                                                  |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IADD:  Push( PopL +  PopR )
        jem#Jem_ISUB:  Push( PopL -  PopR )
        jem#Jem_INEG:  Push(      -  Pop  )
        jem#Jem_IOR:   Push( PopL |  PopR )
        jem#Jem_ISHL:  Push( PopL << PopR )
        jem#Jem_ISHR:  Push( PopL -> PopR )
        jem#Jem_IUSHR: Push( PopL >> PopR )
        jem#Jem_IAND:  Push( PopL &  PopR )
        jem#Jem_IXOR:  Push( PopL ^  PopR )

        ' .---------------------------------------------------------------------------------------------.
        ' |     I2B                                                                                 $3A |
        ' |     Meaning: Integer to signed byte.                                                        |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_I2B:

          if Tos <> SignExtendByte( Tos )
            Crashed( String( "Signed Int too big for Signed Byte" ) )
             
        ' .---------------------------------------------------------------------------------------------.
        ' |     GOTO                                                                                $3B |
        ' |     Meaning: Branch.                                                                        |
        ' |     Description: Execution proceeds at the signed 16-bit offset from the address            |
        ' |     of this instruction. The offset is constructed from byte1 and byte2.                    |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_GOTO:

          BranchIf( true )

        ' .---------------------------------------------------------------------------------------------.
        ' |     INVOKEINTERFACE                                                                     $3C |
        ' |     Meaning: Invoke interface method.                                                       |
        ' |     Note                                                                                    |
        ' |     Unlike invokevirtual and invokespecial, the method block does not indicate the          |
        ' |     number of available arguments. This number is taken from the bytecode.                  |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_INVOKEINTERFACE:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     ARRAYLENGTH                                                                         $3D |
        ' |     Meaning: Get length of array.                                                           |
        ' |     Description: The length of the array is determined and replaces aref on the             |
        ' |     top of the stack. The aref must be a reference to an array object.                      |
        ' |     Note                                                                                    |
        ' |     A NullPointerException is thrown if the aref is null.                                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ARRAYLENGTH:

          PopArrayRef( jem#T_ANY, @arrayPtr, @arrayLen )
          Push( arrayLen)
          
        ' .---------------------------------------------------------------------------------------------.
        ' |     ATHROW                                                                              $3E |
        ' |     Meaning: Throw exception or error.                                                      |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ATHROW:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     CHECKCAST                                                                           $3F |
        ' |     Meaning: Make sure object is of given type.                                             |
        ' |     Description: Determines whether the referenced object can be cast to be a               |
        ' |     reference to an object of another class. A null object reference can be cast to         |
        ' |     any class. Otherwise, the referenced object must be an instance of the expected         |
        ' |     class or one of its superclasses. byte1 and byte2 are used to construct an              |
        ' |     index into the constant pool of the current class, which is presumed to be a            |
        ' |     class name that can be resolved to a class pointer.                                     |
        ' |     Note                                                                                    |
        ' |     A ClassCastException is thrown if the referenced object cannot be cast to the           |
        ' |     expected class.                                                                         |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_CHECKCAST:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     Catch all unimplemented opcodes                                                         |
        ' `---------------------------------------------------------------------------------------------'

        other: NotImplemented

    else
      case opcode

        ' .---------------------------------------------------------------------------------------------.
        ' |     INSTANCEOF                                                                          $40 |
        ' |     Meaning: Determine if an object is of given type and return result.                     |
        ' |     Description: Determines whether the referenced object can be cast to be a               |
        ' |     reference to an object of the expected class. This instruction will overwrite           |
        ' |     the object reference with 1 if the object is an instance of the expected class          |
        ' |     or one of its superclasses. Otherwise, the object reference is overwritten by 0.        |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_INSTANCEOF:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     WIDE                                                                                $41 |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_WIDE:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     IINC                                                                                $42 |
        ' |     Meaning: Increment local variable by constant.                                          |
        ' |     Description: Local variable at byte1 in the current Java frame must contain an          |
        ' |     integer. Its value is incremented by the value byte2, where byte2 is treated as         |
        ' |     a signed 8-bit quantity.                                                                |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IINC:

          IncVar( Operand_UnSignedByte, Operand_SignedByte )

        ' .---------------------------------------------------------------------------------------------.
        ' |     GETFIELD                                                                            $43 |
        ' |     Meaning: Get 16bit field from object.                                                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_GETFIELD:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     PUTFIELD                                                                            $44 |
        ' |     Meaning: Set 16bit field in object.                                                     |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_PUTFIELD:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     NEW                                                                                 $45 |
        ' |     Meaning: Create new object.                                                             |
        ' |     Description: byte1 and byte2 are used to construct an index into the constant           |
        ' |     pool of the current class, which must be a class name that can be resolved to a         |
        ' |     class pointer. A new instance of that class is then created, and a reference to         |
        ' |     the object is pushed on the stack.                                                      |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_NEW:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     NEWARRAY                                                                            $46 |
        ' |     Meaning: Allocate new array.                                                            |
        ' |     Description: A new array of a specific array type, capable of holding size              |
        ' |     elements, is allocated. The result is a reference to the new object. Allocation         |
        ' |     of an array large enough to contain size items of the specific array type is            |
        ' |     attempted and all elements of the array are initialized to 0.                           |
        ' |     size represents the number of elements in the new array and must be an                  |
        ' |     integer. The result is stored with an internal code that indicates the type of          |
        ' |     array to allocate. Possible values for the type of array are as follows:                |
        ' |     T_BOOLEAN(4), T_chAR(5), T_FLOAT(6), T_DOUBLE(7), T_BYTE(8), T_SHORT(9),                |
        ' |     T_INT(10), and T_LONG(11).                                                              |
        ' |     Note                                                                                    |
        ' |     A NegativeArraySizeException is thrown if size is less than 0. An                       |
        ' |     OutOfMemoryError is thrown if there is not enough memory to allocate the array.         |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_NEWARRAY:

          NotImplemented
{                    
          arrayLen := Pop

          Operand_UnsignedByte ' ***** HACK **** This is $00 so what's it mean ?
          
          Push( NewArrayRef( arrayType, arrayLen ) )
}
        ' .---------------------------------------------------------------------------------------------.
        ' |     ANEWARRAY                                                                           $47 |
        ' |     Meaning: Allocate new array of objects.                                                 |
        ' |     Description: A new array of the indicated class type and capable of holding             |
        ' |     size elements is allocated. The result is a reference to the new object.                |
        ' |     Allocation of an array large enough to contain size elements of the given class         |
        ' |     type is attempted and all elements of the array are initialized to null.                |
        ' |     size represents the number of elements in the new array and must be an                  |
        ' |     integer. byte1 and byte2 are used to construct an index into the constant pool          |
        ' |     of the current class. When the item at that index is resolved, the resulting            |
        ' |     entry must be a class.                                                                  |
        ' |     The anewarray instruction is used to create a single-dimension array.                   |
        ' |     Note                                                                                    |
        ' |     A NegativeArraySizeException is thrown if size is less than 0. An                       |
        ' |     OutOfMemoryError is thrown if there is not enough memory to allocate the array.         |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_ANEWARRAY:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     IALOAD                                                                              $48 |
        ' |     Meaning: Load integer from array.                                                       |
        ' |     Description: The integer value at the array index is retrieved and pushed onto          |
        ' |     the top of the stack. The aref must be a reference to an array of integers;             |
        ' |     likewise, the index into the array must be an integer.                                  |
        ' |     Note                                                                                    |
        ' |     A NullPointerException is thrown if aref is null. An                                    |
        ' |     ArrayIndexOutOfBoundsException is thrown if the array index is not within the           |
        ' |     bounds of the array.                                                                    |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IALOAD:

          arrayIndex := Pop
          PopArrayRef( jem#T_INT, @arrayPtr, @arrayLen )

          if arrayIndex < 0 or arrayIndex => arrayLen
            Crashed( String( "Array Index out of bounds" ) )

          if INT32
            push( long[ arrayPtr + ( arrayIndex << 2 ) ] )
          else
            Push( SignExtendWord( word[ arrayPtr + ( arrayIndex << 1 ) ] ) )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IASTORE                                                                             $49 |
        ' |     Meaning: Store into integer array.                                                      |
        ' |     Description: An integer value is popped off the stack and stored in the array           |
        ' |     at the index. The aref must be a reference to an array of integer values, and           |
        ' |     the index into the array must be an integer as well.                                    |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IASTORE:

          value      := Pop
          arrayIndex := Pop
          PopArrayRef( jem#T_INT, @arrayPtr, @arrayLen )

          if arrayIndex < 0 or arrayIndex => arrayLen
            Crashed( String( "Array Index out of bounds" ) )
             
          if INT32
            long[ arrayPtr + ( arrayIndex << 2 ) ] := value
          else
            word[ arrayPtr + ( arrayIndex << 1 ) ] := value

        ' .---------------------------------------------------------------------------------------------.
        ' |     BALOAD                                                                              $4A |
        ' |     Meaning: Load signed byte from array.                                                   |
        ' |     Description: The signed byte value at the array index is retrieved, expanded            |
        ' |     to an integer, and pushed onto the top of the stack. The aref must be a                 |
        ' |     reference to an array of signed byte values, and the index into the array must          |
        ' |     be an integer.                                                                          |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_BALOAD:

          arrayIndex := Pop
          PopArrayRef( jem#T_BYTE, @arrayPtr, @arrayLen )

          if arrayIndex < 0 or arrayIndex => arrayLen
            Crashed( String( "Array Index out of bounds" ) )

          Push( SignExtendByte( byte[ arrayPtr + arrayIndex ] ) )

        ' .---------------------------------------------------------------------------------------------.
        ' |     BASTORE                                                                             $4B |
        ' |     Meaning: Store into signed byte array.                                                  |
        ' |     Description: An integer is popped off the stack, converted to a signed byte,            |
        ' |     and stored in the array at the index. The arrayref must be a reference to an            |
        ' |     array of signed bytes, and the index into the array must be an integer. If the          |
        ' |     integer value is too large to be a signed byte, it is truncated.                        |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_BASTORE:

          value      := Pop
          arrayIndex := Pop
          PopArrayRef( jem#T_BYTE, @arrayPtr, @arrayLen )
          
          if arrayIndex < 0 or arrayIndex => arrayLen
            Crashed( String( "Array Index out of bounds" ) )

          if value <> SignExtendByte( value )
            Crashed( String( "Signed Int too big for Signed Byte" ) )
             
          byte[ arrayPtr + arrayIndex ] := value

        ' .---------------------------------------------------------------------------------------------.
        ' |     LDC                                                                                 $4C |
        ' |     Meaning: Push item from constant pool.                                                  |
        ' |     Description: byte1 is used as an unsigned 8-bit index into the constant pool            |
        ' |     of the current class. After the item at that index is resolved, it is pushed            |
        ' |     onto the stack.                                                                         |
        ' |     Note                                                                                    |
        ' |     This bytecode instruction throws an OutOfMemoryError if a String object is              |
        ' |     being pushed and there is not enough memory to allocate space for it.                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_LDC:

          LoadConstant( 1, Operand_UnsignedByte )

        ' .---------------------------------------------------------------------------------------------.
        ' |     LDC_W                                                                               $4D |
        ' |     Meaning: Push 2-byte item from constant pool.                                           |
        ' |     Description: byte1 and byte2 are used to construct an unsigned 16-bit index             |
        ' |     into the constant pool of the current class. After the item at that index is            |
        ' |     resolved, it is pushed onto the stack.                                                  |
        ' |     Note                                                                                    |
        ' |     This bytecode instruction throws an OutOfMemoryError if a String object is              |
        ' |     being pushed and there is not enough memory to allocate space for it.                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_LDC_W:

          LoadConstant( 2, Operand_UnsignedWord )

        ' .---------------------------------------------------------------------------------------------.
        ' |     LDC2_W                                                                              $4E |
        ' |     Meaning: Push long or double from constant pool.                                        |
        ' |     Description: byte1 and byte2 are used to construct an unsigned 16-bit index             |
        ' |     into the constant pool of the current class. After the two-word constant at             |
        ' |     that index is resolved, it is pushed onto the stack.                                    |
        ' |     Note                                                                                    |
        ' |     This bytecode instruction throws an OutOfMemoryError if a String object is              |
        ' |     being pushed and there is not enough memory to allocate space for it.                   |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_LDC2_W:

          LoadConstant( 4, Operand_UnsignedWord )

        ' .---------------------------------------------------------------------------------------------.
        ' |     IMUL, IDIV, IREM                                                                    $4F |
        ' |     Meaning: Integer maths                                                                  |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_IMUL: Push( PopL *  PopR )
        jem#Jem_IDIV: Push( PopL /  PopR )
        jem#Jem_IREM: Push( PopL // PopR )

        ' .---------------------------------------------------------------------------------------------.
        ' |     JSR                                                                                 $52 |
        ' |     Meaning: Jump subroutine.                                                               |
        ' |     Description: Execution proceeds at the signed 16-bit offset from the address            |
        ' |     of this instruction. The offset is constructed from byte1 and byte2. The                |
        ' |     address of the instruction immediately following the current instruction is             |
        ' |     pushed onto the stack to provide a return address from the subroutine.                  |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_JSR:

          Push( pc )
          BranchIf( true )

        ' .---------------------------------------------------------------------------------------------.
        ' |     RET                                                                                 $53 |
        ' |     Meaning: Return from subroutine.                                                        |
        ' |     Description: Local variable in the current Java frame must contain a return             |
        ' |     address. The contents of the local variable are written into the program                |
        ' |     counter.                                                                                |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_RET:

          pc := GetVar( Operand_UnsignedByte )

        ' .---------------------------------------------------------------------------------------------.
        ' |     TABLESWITCH                                                                         $54 |
        ' |     Meaning: Access jump table by index and jump.                                           |
        ' |     Description: Immediately after the tableswitch instruction, padding consisting          |
        ' |     of 0-3 zeros is inserted so that the next byte begins at an address that is a           |
        ' |     multiple of four. A series of offsets follows the padding. These offsets are            |
        ' |     signed 4-byte quantities and consist of a default offset, a low offset, and a           |
        ' |     high offset, followed by additional high - low + 1 offsets. These additional            |
        ' |     offsets are treated as a 0-based jump table.                                            |
        ' |     The index the offsets point to must be an integer. If the index is less than            |
        ' |     the low offset or greater than the high offset, the default offset is added to          |
        ' |     the address of the current instruction. Otherwise, the low offset is subtracted         |
        ' |     from the index, and the element at the position index - low offset in the jump          |
        ' |     table is extracted and added to the address of the current instruction.                 |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_TABLESWITCH:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     LOOKUPSWITCH                                                                        $55 |
        ' |     Meaning: Access jump table by key match and jump.                                       |
        ' |     Description: Immediately after the lookupswitch instruction, padding                    |
        ' |     consisting of 0-3 zeros is                                                              |
        ' |     inserted so that the next byte begins at an address that is a multiple of               |
        ' |     four. A series of pairs of offsets follows the padding. These offsets are               |
        ' |     signed 4-byte quantities. The first item in the pair is the default offset, and         |
        ' |     the second item in the pair gives the number of pairs that follow. The                  |
        ' |     additional pairs consist of a match and an offset.                                      |
        ' |     The integer key on the stack is then compared against each of the matches. If           |
        ' |     the key is equal to one of the matches, the offset is added to the address of           |
        ' |     the current instruction. If the key does not match any of the matches, the              |
        ' |     default offset is added to the address of the current instruction.                      |
        ' `---------------------------------------------------------------------------------------------'

        jem#Jem_LOOKUPSWITCH:

          NotImplemented

        ' .---------------------------------------------------------------------------------------------.
        ' |     Catch all unimplemented opcodes                                                         |
        ' `---------------------------------------------------------------------------------------------'

        other: NotImplemented

' .-----------------------------------------------------------------------------------------------------.
' |     Native Method Handling                                                                          |
' `-----------------------------------------------------------------------------------------------------'

PRI NativeMethod( nativeMethodNumber, numberOfParameters ) | i, rc

  ' On entry the stack will hold the numberOfparameters needed to execute the native method, with BP 
  ' set as pointing to the first parameter, SP pointing to the last parameter. Pop() will remove in
  ' reverse order from SP down to BP. The stack can be pushed and popped at will during the Native
  ' Method handler so long as it is not overflowed or underflowed.
  '
  ' When leaving the Native Method handler any return result should have been pushed to the stack and
  ' a "result := true" issued to indicate there is a value returned, otherwise 'result' should not be
  ' set and the Native Method handler simply exited.
  '
  ' For example a native method call to "stamp.core.CPU.Count(A,B,C)" will deliver a numberOfParameters
  ' count of 3 and a stack as follows -
  ' 
  '       .---. --.         
  ' BP -->| A |   |
  '       | B |   |-- Number of Parameters = 3
  ' SP -->| C |   |
  '       `---' --'

  opcode := -1

  if natDebug
    tv.Str( String( $0D,"Native       " ) )
    tv.Hex( nativeMethodNumber, 2 )
    tv.Out( " " )
    tv.Str( jem.nativeMethodNamePtr( nativeMethodNumber ) )
    tv.Out( "(" )
    if numberOfParameters
      repeat i from 0 to numberOfParameters-1
        if i <> 0
          tv.Out( "," )
        tv.Hex( GetVar( i ), 4 )
    tv.Out( ")" )

  ' do work here
  rc := do_native( nativeMethodNumber, numberOfParameters )
  
  ' set return here  
  case nativeMethodNumber

    jem#Native_CPU_writeRegister:

      NotImplemented

    jem#Native_CPU_readRegister:

      NotImplemented
      Push(rc)
      result := true
      
    jem#Native_CPU_writeObject:

      NotImplemented

    jem#Native_CPU_delay:

      NotImplemented

    jem#Native_CPU_message:

      NotImplemented

    jem#Native_CPU_count:

      NotImplemented
      Push(rc)
      result := true

    jem#Native_CPU_rcTime:

      NotImplemented
      Push(rc)
      result := true

    jem#Native_CPU_pulseIn:

      NotImplemented
      Push(rc)
      result := true

    jem#Native_CPU_pulseOut:

      NotImplemented

    jem#Native_CPU_readPin:

      NotImplemented
      Push(rc)
      result := true

    jem#Native_CPU_writePin:

      NotImplemented

    jem#Native_CPU_shiftIn:

      NotImplemented
      Push(rc)
      result := true

    jem#Native_CPU_shiftOut:

      NotImplemented

    jem#Native_CPU_carry:

      NotImplemented
      Push(rc)
      result := true

    jem#Native_CPU_uninstall:

      NotImplemented

    jem#Native_CPU_install:

      NotImplemented

    jem#Native_CPU_nap:

      NotImplemented

    jem#Native_CPU_setOutput:

      NotImplemented

    jem#Native_CPU_setInput:

      NotImplemented

    jem#Native_CPU_readSRAM:

      NotImplemented
      Push(rc)
      result := true

    jem#Native_CPU_writeSRAM:

      NotImplemented

    jem#Native_EEPROM_eeRead:

      NotImplemented
      Push(rc)
      result := true

    jem#Native_EEPROM_eeWrite:

      NotImplemented

    jem#Native_Uart_txInit:

      NotImplemented

    jem#Native_Uart_rxInit:

      NotImplemented

    jem#Native_Uart_rxRead:

      NotImplemented
      Push(rc)
      result := true

    jem#Native_Callback_returnTwo:

      NotImplemented
      Push(rc)
      result := true

    jem#Native_Timer_init:

      NotImplemented

    jem#Native_Timer_latch:

      NotImplemented

    jem#Native_PWM_install:

      NotImplemented

    jem#Native_PWM_updateInternal:

      NotImplemented

    jem#Native_DAC_install:

      NotImplemented

    jem#Native_ADC_install:

      NotImplemented

    jem#Native_Terminal_getByte:

      NotImplemented
      Push(rc)
      result := true

    other:
    
      NotImplemented


' .-----------------------------------------------------------------------------------------------------.
' |     Deal with unimplemened opcodes and forced PASM calls                                            |
' `-----------------------------------------------------------------------------------------------------'

PRI NotImplemented

  if opcode < 0
    if not allowNotImplementedNatives
      Crashed( String( "Native Method not implemented" ) )
  else
    if mode == MODE_SPIN_ONLY
      if doJazzLib == 0
      else
        tv.Str( String( $0D,"PC " ) )
        tv.Hex( pc-bytecodeStart, 2 )
        tv.Str( String( " OP " ) )
        tv.Hex( byte[pc-bytecodeStart], 2 )
        Crashed( String( "Spin did not handle opcode" ) )
    else
      if mode == MODE_SPIN_THEN_PASM
        if JvmPasm
          abort
        else
          Crashed( String( "Spin then PASM did not handle opcode" ) )
      else
        if doJazzLib == 0
        else
          Crashed( String( "PASM then Spin did not handle opcode" ) )

PRI UsePasm

  if JvmPasm
    abort
  else
    Crashed( String( "Forced PASM did not handle opcode" ) )

' .-----------------------------------------------------------------------------------------------------.
' |     Bytecode Operand Handling                                                                       |
' `-----------------------------------------------------------------------------------------------------'

PRI Operand_UnSignedByte : operand

  operand := byte[ pc++ ]

  if debugging
    tv.Out( " " )
    tv.Out( "$" )
    tv.Hex( operand, 2 )
  
PRI Operand_SignedByte : operand

  operand := SignExtendByte( Operand_UnSignedByte )

PRI Operand_UnSignedWord : operand

  operand := byte[ pc++ ] << 8
  operand |= byte[ pc++ ]

  if debugging
    tv.Out( " " )
    tv.Out( "$" )
    tv.Hex( operand,4)
  
PRI Operand_SignedWord : operand

  operand := SignExtendWord( Operand_UnSignedWord )

PRI Operand_BytecodeAddress : operand

  operand := Operand_UnSignedWord + bytecodeStart

PRI SignExtendByte( n )

  result := ~ n

PRI SignExtendWord( n )

  result := ~~ n


' .-----------------------------------------------------------------------------------------------------.
' |     Stack Pushing                                                                                   |
' `-----------------------------------------------------------------------------------------------------'

PRI Push( n )

  if INT32
    PushLong( n )
  else
    PushWord( n )
    
PRI PushLong( n )

  stackAltered := true
  if INT32
    sp += 4
    if sp > @stack + ( STACK_SIZE << INTWORDS )
      Crashed( String( "Stack overflow" ) )
    long[ sp ] := n
  else
    Crashed( String( "Cannot use PushLong if not INT32" ) )
   
PRI PushWord( n )

  stackAltered := true
  sp += 2
  if sp > @stack + ( STACK_SIZE << INTWORDS )
    Crashed( String( "Stack overflow" ) )
  word[ sp ] := n

PRI PushAddr( n )

  if INT32
    PushLong( n )
  else
    if ChipVer == 1
      PushWord( n )
    else
      PushWord( n >> 16 )
      PushWord( n       )
      
' .-----------------------------------------------------------------------------------------------------.
' |     Stack Popping                                                                                   |
' `-----------------------------------------------------------------------------------------------------'

PRI Pop

  if INT32
    result := PopLong
  else
    result := PopWord

PRI PopLong

  stackAltered := true
  if INT32
    if sp < @stack
      Crashed( String( "Stack underflow" ) )
    result := long[ sp ]
    sp -= 4
  else
    Crashed( String( "Cannot use PopLong if not INT32" ) )
    
PRI PopWord

  stackAltered := true
  if sp < @stack
    Crashed( String( "Stack underflow" ) )
  result := SignExtendWord( word[ sp ] )
  sp -= 2

PRI PopAddr

  if INT32
    result := PopLong
  else
    if ChipVer == 1
      result := PopWord
    else
      result := PopWord
      result |= PopWord << 16

PRI Tos

  result := Pop
  Push( result )
        
' .-----------------------------------------------------------------------------------------------------.
' |     Stack Manipulation                                                                              |
' `-----------------------------------------------------------------------------------------------------'

PRI PopL : lhs | rhs

  '       .---.           .---.
  '       | L |     SP -->| R |
  '       |---|           `---'
  ' SP -->| R |
  '       `---'

  rhs := Pop
  lhs := Pop
  Push( rhs )

PRI PopR : rhs

  rhs := Pop

PRI Pop2

  Pop
  Pop

PRI Dup | a

  '       .---.           .---.
  ' SP -->| A |           | A |
  '       `---'           |---|
  '                 SP -->| A |
  '                       `---'

  a := Pop

  Push( a )
  Push( a )

PRI Dup_X1 | a, b

  '       .---.           .---.
  '       | B |           | A |
  ' SP -->| A |           |---|
  '       `---'           | B |
  '                 SP -->| A |
  '                       `---'
  a := Pop
  b := Pop

  Push( a )
  Push( b )
  Push( a )

PRI Dup_X2 | a, b, c

  '       .---.           .---.
  '       | C |           | A |
  '       | B |           |---|
  ' SP -->| A |           | C |
  '       `---'           | B |
  '                 SP -->| A |
  '                       `---'
  a := Pop
  b := Pop
  c := Pop

  Push( a )
  Push( c )
  Push( b )
  Push( a )

PRI Dup2 | a, b

  '       .---.           .---.
  '       | B |           | B |
  ' SP -->| A |           | A |
  '       `---'           |---|
  '                       | B |
  '                 SP -->| A |
  '                       `---'

  a := Pop
  b := Pop

  Push( b )
  Push( a )
  Push( b )
  Push( a )

PRI Dup2_X1 | a, b, c

'       .---.           .---.
'       | C |           | B |
'       | B |           | A |
' SP -->| A |           |---|
'       `---'           | C |
'                       | B |
'                 SP -->| A |
'                       `---'

  a := Pop
  b := Pop
  c := Pop

  Push( b )
  Push( a )
  Push( c )
  Push( b )
  Push( a )

PRI Dup2_X2 | a, b, c, d

  '       .---.           .---.
  '       | D |           | B |
  '       | C |           | A |
  '       | B |           |---|
  ' SP -->| A |           | D |
  '       `---'           | C |
  '                       | B |
  '                 SP -->| A |
  '                       `---'

  a := Pop
  b := Pop
  c := Pop
  d := Pop

  Push( b )
  Push( a )
  Push( d )
  Push( c )
  Push( b )
  Push( a )

PRI Swap | a, b

  '       .---.           .---.
  '       | B |           | A |
  ' SP -->| A |     SP -->| B |
  '       `---'           `---'

  a := Pop
  b := Pop

  Push( a )
  Push( b )

' .-----------------------------------------------------------------------------------------------------.
' |    Variable Handling                                                                                |
' `-----------------------------------------------------------------------------------------------------'

PRI GetVar( offset )

  result := Peek( bp, offset )

PRI PutVar( offset, n )

  Poke( bp, offset, n )

PRI IncVar( offset, n )

  PutVar( offset, GetVar( offset ) + n )
  
PRI Peek( base, offset )

  if INT32
    result := long[ base + ( offset << 2 ) ]
  else
    result := SignExtendWord( word[ base + ( offset << 1 ) ] )
{
  if 0'jazzDebug
    tv.Str( string($d,"Peek "))
    tv.Hex( base, 4 )
    tv.Out( " ")
    tv.Hex( offset, 4 )
    tv.Str( string(" rc "))
    tv.Hex( result, 4 )
}
PRI Poke( base, offset, n )

  if INT32
    long[ base + ( offset << 2 ) ] := n
  else
    word[ base + ( offset << 1 ) ] := n
{
  if 0'jazzDebug
    tv.Str( string($d,"Poke "))
    tv.Hex( base, 4 )
    tv.Out( " ")
    tv.Hex( offset, 4 )
    tv.Str( string(" val "))
    tv.Hex( n, 4 )
}
' .-----------------------------------------------------------------------------------------------------.
' |    Arrays                                                                                           |
' `-----------------------------------------------------------------------------------------------------'
'
' How an array is held in memory
' .------.--------.---.---.- - -.---.
' | Type | BytLen | 0 | 1 |     | N |
' `------^--------^---^---^- - -^---'
'
' An Array Reference is a pointer to the Type
' The Type is an INT that holds a value of jem#T_XXXX
' The BytLen is the number of bytes used to hold the array data
' A returned Array Start is a pointer to the first data item
' The Type and Bytlen are always Int(16) or Int(32) depending upon ChipVer *NOT* INT32 
' The array data entity size depends upon Typ

PRI NewArrayRef( typ, len )

  result := GetFromHeap( typ, len )
    
PRI PopArrayRef( typ, atPtr, atLen ) | arrayTyp, bytLen, datPtr

  result := Pop
{
  if jazzDebug
    tv.Str(string($d,"PopArrayRef "))
    tv.Str(jem.TypeNamePtr(typ))
    tv.Str(string(" refptr "))
    tv.Hex(result, 4)
    tv.Out(" ")
}
  if result == 0
    Crashed( String( "Got Null Pointer for Array Ref" ) )
  else
     
    GetHeapInfo( result, @arrayTyp, @bytLen, @datPtr )
{
    if jazzDebug
      tv.Str(string(" type "))
      tv.Hex(arrayTyp, 4)
}
    long[ atPtr ] := datPtr
    long[ atLen ] := bytLen >> BytesShiftPerType( arrayTyp )
{
    if jazzDebug
      tv.Str(string(" array "))
      tv.Hex(word[atPtr], 4)
      tv.Str(string(" len "))
      tv.Hex(word[atLen], 4)
}
    if arrayTyp <> typ
      if typ <> jem#T_ANY
        Crashed( String( "Incorrect Array Type" ) )
     
' .-----------------------------------------------------------------------------------------------------.
' |    Objects                                                                                          |
' `-----------------------------------------------------------------------------------------------------'

PRI PopObjectRef

  NotImplemented
  
' .-----------------------------------------------------------------------------------------------------.
' |    Heap Management                                                                                  |
' `-----------------------------------------------------------------------------------------------------'
'
' How an array is held in memory
' .------.--------.---.---.- - -.---.
' | Type | BytLen | 0 | 1 |     | N |
' `------^--------^---^---^- - -^---'
'
' An Array Reference is a pointer to the Type
' The Type is an INT that holds a value of jem#T_XXXX
' The BytLen is the number of bytes used to hold the array data
' A returned Array Start is a pointer to the first byte of data
' The Type and Bytlen are always Int(16) or Int(32) depending upon INT32
' The array data entity size depends upon Typ

PRI GetFromHeap( typ, len ) | bytLen, i

  bytLen := len << BytesShiftPerType(typ)
{  
  if jazzDebug
    tv.Out($0D)
    tv.Str( String( "GetFromHeap Typ  " ) )
    tv.Hex( typ,4 )
    tv.Out( " ")
    tv.Str( jem.TypeNamePtr(typ) )
    tv.Str( String( " Len  " ) )
    tv.Hex( len, 4 )
    tv.Str( String( " BytLen " ) )
    tv.Hex( bytLen, 4 )
}
  ' Pause_Ms( 10_000 )

  if ( ChipVer == 1 ) or ( not BIGARRAYS )
    heap &= $FFFF_FFFE
    heap -= bytLen + ( HEAP_HEAD_DATA << HEAP_WORD )
    word[ heap + ( HEAP_HEAD_TYPE << HEAP_WORD ) ] := typ
    word[ heap + ( HEAP_HEAD_BLEN << HEAP_WORD ) ] := bytLen
    i  := heap + ( HEAP_HEAD_DATA << HEAP_WORD )
  else
    heap &= $FFFF_FFFC
    heap -= bytLen + ( HEAP_HEAD_DATA << HEAP_LONG )
    long[ heap + ( HEAP_HEAD_TYPE << HEAP_LONG ) ] := typ
    long[ heap + ( HEAP_HEAD_BLEN << HEAP_LONG ) ] := bytLen
    i  := heap + ( HEAP_HEAD_DATA << HEAP_LONG )

  repeat bytLen
    byte[ i++ ] := 0
{
  if jazzDebug
    tv.Str(string(" heap "))
    tv.Hex(heap, 4)
}
  return heap

PRI GetHeapInfo( ptr, atTyp, atLen, atDat ) : nextPtr | typ, Bytlen, datPtr, ii, ts
 
  if ( ChipVer == 1 ) or ( not BIGARRAYS )
    typ    := word[ ptr + ( HEAP_HEAD_TYPE << HEAP_WORD ) ]
    bytLen := word[ ptr + ( HEAP_HEAD_BLEN << HEAP_WORD ) ]
    datPtr := ptr + ( HEAP_HEAD_DATA << HEAP_WORD )
    nextPtr := datPtr + bytLen
    if nextPtr & 1
      nextPtr++
  else
    typ    := long[ ptr + ( HEAP_HEAD_TYPE << HEAP_LONG ) ]
    bytLen := long[ ptr + ( HEAP_HEAD_BLEN << HEAP_LONG ) ]
    datPtr := ptr + ( HEAP_HEAD_DATA << HEAP_LONG )
    nextPtr := datPtr + bytLen
    if nextPtr & %11
      nextPtr := ( nextPtr | %11 ) + 1

  if atTyp
    long[ atTyp ] := typ
  if atLen
    long[ atLen ] := bytLen
  if atdat
    long[ atDat ] := datPtr

  ts := jem.TypeNamePtr(typ) 
  if byte[ts] <> "U"
    nextPtr := HEAP_BASE 'let ShowHeap know to stop
{    
  if 0'jazzDebug
    tv.Out($0D)
    tv.Str(String( "GetHeapInfo  " ))
    tv.Hex(ptr,4)
    tv.Out(" ")
    ts := jem.TypeNamePtr(typ) 
    tv.Str(ts)
    if byte[ts] <> "U"
      tv.Str(String(" Len  "))
      if typ == jem#T_BYTE
        tv.Dec(bytLen)
      else
        tv.Dec(bytLen/2)
      tv.Str(String(" Dat "))
      tv.hex(datPtr,4)
      repeat ii from 0 to bytLen
        if typ == jem#T_BYTE
          if (ii // 16) == 0
            tv.out($d)
          tv.hex(byte[datPtr+ii],2)
        else         
          if (ii // 32) == 0
            tv.out($d)
          tv.hex(word[datPtr+ii++],4)
        tv.out(" ")
    else
      ShowHeap
      Crashed(string($D,"GetHeapInfo lookup failed.",$D))
}      
      
PRI BytesShiftPerType( typ )

  case typ & jem#T_ANY
  
    jem#T_BOOLEAN: result := ( >| 1        ) - 1
    jem#T_CHAR:    result := ( >| CHRBYTES ) - 1
    jem#T_FLOAT:   result := ( >| 2        ) - 1
    jem#T_DOUBLE:  result := ( >| 2        ) - 1
    jem#T_BYTE:    result := ( >| 1        ) - 1
    jem#T_SHORT:   result := ( >| 2        ) - 1
    jem#T_INT:     result := ( >| INTBYTES ) - 1
    jem#T_LONG:    result := ( >| LNGBYTES ) - 1
    jem#T_ANY:     result := ( >| INTBYTES ) - 1
    
    other:
      tv.out($D)
      tv.hex(typ,4)
      Crashed( String( " BytesShiftPerType Unknown Type" ) )

PRI BytesPerType( typ )

  case typ & jem#T_ANY
  
    jem#T_BOOLEAN: result := 1
    jem#T_CHAR:    result := 2
    jem#T_FLOAT:   result := 4
    jem#T_DOUBLE:  result := 4
    jem#T_BYTE:    result := 1
    jem#T_SHORT:   result := 2
    jem#T_INT:     result := 2
    jem#T_LONG:    result := 4
    jem#T_ANY:     result := 4
    
    other:
      tv.out($D)
      tv.hex(typ,4)
      Crashed( String( " BytesPerType Unknown Type" ) )

PRI TypePerSize( size )

  case size
    1: result := jem#T_BYTE
    2: result := jem#T_SHORT
    4: result := jem#T_INT 
    8: result := jem#T_LONG
    other: result := 2

PRI GetHeapValuePtr(ptr, index) | type, len, data
  GetHeapInfo(ptr, @type, @len, @data)
  return GetHeapDataPtr(data, type, len, index)

PRI GetHeapValue(ptr, index) | type, len, data, shift
{{
/**
 * get value from heap pointer[index]
 * @param ptr - the heap pointer
 * @param index - the heap address offset data index
 * @returns value at ptr[index]
 */
GetHeapValue(ptr, index);
}}
  GetHeapInfo(ptr, @type, @len, @data)
  if (index => 0) and (index < len<<BytesShiftPerType(type))
    return GetHeapDataValue(data, type, len, index)
  else
    tv.out($d)
    tv.dec(index)
    Crashed(string(" GetHeapValue index out of range"))


PRI SetHeapValue(ptr, index, val) | type, len, data, shift
{{
/**
 * set value at heap pointer[index]
 * @param ptr - the heap pointer
 * @param index - the heap address offset data index
 * @param value - value to set at ptr[index]
 */
SetHeapValue(ptr, index, val);
}}
  GetHeapInfo(ptr, @type, @len, @data)
  if (index => 0) and (index < len<<BytesShiftPerType(type))
    SetHeapDataValue(data, type, len, index, val)
  else
    tv.out($d)
    tv.dec(index)
    tv.out(">")
    tv.dec(len)
    Crashed(string(" GetHeapValue index out of range"))

PRI GetHeapDataPtr(dataptr, type, len, index) | shift
{{
/**
 * get value from heap data[index]
 * @note dataptr, type, & len must be previously retrieved with GetHeapInfo
 *
 * @param dataptr - the heap data pointer
 * @param type - the type stored
 * @param len - the number of bytes in this entry
 * @param index - the heap dataptr address offset index
 * @returns value at ptr[index]
 */
GetHeapDataValue(dataptr, type, len, index);
}}
  if (index => 0) and (index < len<<BytesShiftPerType(type))
    shift := BytesShiftPerType(type)
    case type
      jem#T_BOOLEAN: result := dataptr+index<<shift
      jem#T_CHAR:    result := dataptr+index<<shift
      jem#T_FLOAT:   result := dataptr+index<<shift
      jem#T_DOUBLE:  result := dataptr+index<<shift
      jem#T_BYTE:    result := dataptr+index<<shift
      jem#T_SHORT:   result := dataptr+index<<shift
      jem#T_INT:     result := dataptr+index<<shift
      jem#T_LONG:    result := dataptr+index<<shift
      jem#T_ANY:     result := dataptr+index<<shift
{
    if 0'jazzDebug
      shift := BytesPerType(type)
      tv.out($d)
      tv.str(string("GetHeapDatVal "))
      tv.hex(dataptr,4)
      tv.out(" ")
      tv.str(jem.TypeNamePtr(type))
      tv.str(string(" ["))
      tv.dec(index)
      tv.str(string("] $"))
      tv.hex(result,shift<<1)
}
  else
    tv.out($d)
    tv.dec(index)
    Crashed(string(" GetHeapDataValue index out of range"))

PRI GetHeapDataValue(dataptr, type, len, index) | shift
{{
/**
 * get value from heap data[index]
 * @note dataptr, type, & len must be previously retrieved with GetHeapInfo
 *
 * @param dataptr - the heap data pointer
 * @param type - the type stored
 * @param len - the number of bytes in this entry
 * @param index - the heap dataptr address offset index
 * @returns value at ptr[index]
 */
GetHeapDataValue(dataptr, type, len, index);
}}
  if (index => 0) and (index < len<<BytesShiftPerType(type))
    shift := BytesShiftPerType(type)
    case type
      jem#T_BOOLEAN: result := word[dataptr+index<<shift]
      jem#T_CHAR:    result := word[dataptr+index<<shift]
      jem#T_FLOAT:   result := word[dataptr+index<<shift]
      jem#T_DOUBLE:  result := long[dataptr+index<<shift]
      jem#T_BYTE:    result := byte[dataptr+index<<shift]
      jem#T_SHORT:   result := word[dataptr+index<<shift]
      jem#T_INT:     result := word[dataptr+index<<shift]
      jem#T_LONG:    result := long[dataptr+index<<shift]
      jem#T_ANY:     result := word[dataptr+index<<shift]
{
    if 0'jazzDebug
      shift := BytesPerType(type)
      tv.out($d)
      tv.str(string("GetHeapDatVal "))
      tv.hex(dataptr,4)
      tv.out(" ")
      tv.str(jem.TypeNamePtr(type))
      tv.str(string(" ["))
      tv.dec(index)
      tv.str(string("] $"))
      tv.hex(result,shift<<1)
}
  else
    tv.out($d)
    tv.dec(index)
    Crashed(string(" GetHeapDataValue index out of range"))

PRI SetHeapDataValue(dataptr, type, len, index, val) | shift
{{
/**
 * set value at heap dataptr[index]
 * @note dataptr, type, & len must be previously retrieved with GetHeapInfo
 *
 * @param dataptr - the heap data pointer
 * @param type - the type stored
 * @param len - the number of bytes in this entry
 * @param index - the heap dataptr address offset index
 * @param value - the value to set at dataptr[index]
 */
SetHeapDataValue(dataptr, type, len, index, val);
}}
  if (index => 0) and (index < len<<BytesShiftPerType(type))
    shift := BytesShiftPerType(type)
    case type
      jem#T_BOOLEAN: word[dataptr+index<<shift] := val
      jem#T_CHAR:    word[dataptr+index<<shift] := val
      jem#T_FLOAT:   word[dataptr+index<<shift] := val
      jem#T_DOUBLE:  long[dataptr+index<<shift] := val
      jem#T_BYTE:    byte[dataptr+index<<shift] := val
      jem#T_SHORT:   word[dataptr+index<<shift] := val
      jem#T_INT:     word[dataptr+index<<shift] := val
      jem#T_LONG:    long[dataptr+index<<shift] := val
      jem#T_ANY:     word[dataptr+index<<shift] := val
{
    if 0'jazzDebug
      shift := BytesPerType(type)
      tv.out($d)
      tv.str(string("SetHeapDatVal "))
      tv.hex(dataptr,4)
      tv.out(" ")
      tv.str(jem.TypeNamePtr(type))
      tv.str(string(" ["))
      tv.dec(index)
      tv.str(string("] $"))
      tv.hex(val,shift<<1)
}
  else
    tv.out($d)
    tv.dec(index)
    Crashed(string(" SetHeapDataValue index out of range"))
         
' .-----------------------------------------------------------------------------------------------------.
' |    Constant Handling                                                                                |
' `-----------------------------------------------------------------------------------------------------'

PRI oldLoadConstant( constantSizeInBytes, tableIndex ) | constAdr, constLen, constOff, heapPtr, heapData, ii

  '              ; String index table.
  ' 0B25: dw $0B ; String offset hi = $0B
  ' 0B26: dw $2B ;               lo = $2B
  ' 0B27: dw $0B ; String offset hi = $0B
  ' 0B28: dw $35 ;               lo = $35
  ' 0B29: dw $0B ; String offset hi = $0B
  ' 0B2A: dw $43 ;               lo = $43

  constAdr := byte[ ct + tableIndex << 1 + 0 ] << 8
  constAdr |= byte[ ct + tableIndex << 1 + 1 ] 
  constAdr += bytecodeStart 
  
  ' 0B35: dw $01 ; String class  hi = $01
  ' 0B36: dw $C9 ;               lo = $C9
  ' 0B37: dw $00 ; String length hi = $00
  ' 0B38: dw $04 ;               lo = $04
  ' 0B39: dw $0B ; Pointer       hi = $0B
  ' 0B3A: dw $3B ;               lo = $3B
  ' 0B3B: dw $01 ; Object class  hi = $01
  ' 0B3C: dw $20 ;               lo = $20
  ' 0B3D: dw $00 ; String length hi = $00
  ' 0B3E: dw $04 ;               lo = $04
  ' 0B3F: dw "n" ; String Data   "null"
  ' 0B40: dw "u"
  ' 0B41: dw "l"
  ' 0B42: dw "l"

  mode := 0 ' We messed with stuff so don't let PASM try !
  NotImplemented

' .-----------------------------------------------------------------------------------------------------.
' |    Flow Control                                                                                     |
' `-----------------------------------------------------------------------------------------------------'

PRI BranchIf ( condition ) | offset

  offset := Operand_SignedWord - 3

  if debugging
    tv.Out( " " )
    tv.Out( "(" )
    tv.Out( " " )
    tv.Dec( pc + offset - mb )
    tv.Out( " " )
    tv.Out( ")" )
  
  if condition
    pc += offset

PRI InvokeProgram( bytecodeStartPtr ) |  constantTable, mainMethodAdr, debstate, numStaticFields, mainMethodBase
 
  bytecodeStart := bytecodeStartPtr
  
  ' 0000: dw $A4 ; Placeholder for run/debug.
  ' 0001: dw $0  ; Placeholder for the length.
  ' 0002: dw $01 ; Version: 0.1
  ' 0003: dw 16  ; Number of static fields = 8
  '
  ' 0004: dw $00 ; `main' method: Offset hi = $00
  ' 0005: dw $1E ;                       lo = $1E
  '
  ' 0006: dw $01 ; java.lang.Object class: Offset hi = $01
  ' 0007: dw $12 ;                                lo = $12
  '
  ' 0008: dw $01 ; java.lang.String class: Offset hi = $01
  ' 0009: dw $BB ;                                lo = $BB
  '
  ' 000A: dw $00 ; Throwable.throwVMException method: Offset hi = $00
  ' 000B: dw $36 ;                                           lo = $36
  '
  ' 000C: dw $0B ; String table offset hi = $0B
  ' 000D: dw $03 ;                     lo = $03
  '
  ' 000E: dw $00 ; Callback method offset hi = $00
  ' 000F: dw $00 ;                        lo = $00
  '
  ' 0010: dw $01 ; java.lang.OutOfMemoryError class: Offset hi = $01
  ' 0011: dw $95 ;                                          lo = $95
  '
  ' 0012: dw $0B ; Stack base hi = $0B
  ' 0013: dw $40 ;            lo = $40

  case byte[ bytecodeStart + $0002 ] ' File Version

    ' Standard Javelin Int(16), no Long
    $01   : if INT32
              tv.Str( String( $0D,"Jem File INT16's auto-promoted to INT32" ) )
            jemFileCountAdjust := 1
            
'== ' Int(16), Long(32)
'== $80   :  if INT32
'==           tv.Str( String( $0D,"Jem File INT16's auto-promoted to INT32" ) )
    
'== ' Int(32), Long(32)
'== $C0   : if not INT32
'==           Crashed( String( "Jem File requires INT32" ) )
            
    other : Crashed( String( "Unknown Jem File Version" ) )

  'tv.Str(String($d,"bytecodeStart $"))
  'tv.hex(bytecodeStart,4)

  gBytecodeLen    := byte[ bytecodeStart + $0001 ] << 8

  'tv.Str(String($d,"bytecodeLen $"))
  'tv.hex(gBytecodeLen,4)

  numStaticFields := byte[ bytecodeStart + $0003 ] >> jemFileCountAdjust

  'tv.Str(String($d,"numStaticFields $"))
  'tv.hex(numStaticFields,4)
   
  mainMethodAdr  := byte[ bytecodeStart + $0004 ] << 8
  mainMethodAdr  |= byte[ bytecodeStart + $0005 ]
  'gBytecodeMain  := mainMethodAdr 

  'tv.Str(String($d,"mainMethodAdr $"))
  'tv.hex(mainMethodAdr,4)
  
  gObjectClassId := byte[ bytecodeStart + $0006 ] << 8
  gObjectClassId |= byte[ bytecodeStart + $0007 ] 
  
  'tv.Str(String($d,"gObjectClassId $"))
  'tv.hex(gObjectClassId,4)
  
  constantTable  := byte[ bytecodeStart + $000C ] << 8
  constantTable  |= byte[ bytecodeStart + $000D ] 

  'tv.Str(String($d,"constantTable $"))
  'tv.hex(constantTable,4)
  
  mainMethodBase := byte[ mainMethodAdr + bytecodeStart + 3] << 8
  mainMethodBase |= byte[ mainMethodAdr + bytecodeStart + 4]

  gBytecodeMain  := mainMethodBase 

  'tv.Str(String($d,"gBytecodeMain $"))
  'tv.hex(gBytecodeMain,4)
  

    
  mainMethodBase += bytecodeStart

  debstate := jazzDebug
  jazzDebug := 0  
  LoadClassInfo(mainMethodBase)
  
{ LoadClassInfo accessor function tests
  LookupClassIdIndex($118)
  LookupMethodIndex($118,$11B)
  LookupMethodClassId($11B)
}
  jazzDebug := debstate

  ' Create a chunk which we can hold our static fields in, and point to the fields ...

  if ( ChipVer == 1 ) or ( not BIGARRAYS )
    vp := GetFromHeap( jem#T_INT, numStaticFields) + ( HEAP_HEAD_DATA << HEAP_WORD )
  else
    vp := GetFromHeap( jem#T_INT, numStaticFields) + ( HEAP_HEAD_DATA << HEAP_LONG )
  
  ct := constantTable + bytecodeStart

  InitConstantTable
  'gMainObject := GetNewClass($0014)

   
  CallMethod( mainMethodAdr + bytecodeStart, false )

    
PRI InvokeStatic( methodBlockAdr )

  CallMethod( methodBlockAdr, false )

  
PRI CallClass( classBlockAdr ) | superClassAdr, nofClassFields 

  '              ; Class = `java/lang/String'
  ' 01BB: dw $01 ; 0 - Superclass offset hi = $01
  ' 01BC: dw $12 ; 1 - Superclass offset lo = $12
  ' 01BD: dw 4   ; 2 - Number of (non-final, non-static) fields = 2

  superClassAdr  := byte[ classBlockAdr + 0 ] << 8
  superClassAdr  |= byte[ classBlockAdr + 1 ] 
  nofClassFields := byte[ classBlockAdr + 2 ] >> jemFileCountAdjust

  Crashed( String( "CallClass not written yet" ) )
  
PRI CallMethod( methodBlockAdr, extra ) | classid, frameStackSize, numberOfParameters, numberOfLocals, flags, methodAdr, exceptionTable, i, oldFramePointer

  '              ; Method = 'MyMethod'
  ' 004E: dw 4   ; 0 - Number of parameters = 2
  ' 004F: dw 6   ; 1 - Number of locals = 3
  ' 0050: dw $00 ; 2 - Flags     = $00
  ' 0051: dw $04 ; 3 - Offset hi = $04
  ' 0052: dw $E8 ; 4 - Offset lo = $E8
  ' 0053: dw $00 ; 5 - Empty exception table hi = $00
  ' 0054: dw $00 ; 6 - Empty exception table lo = $00

  numberOfParameters := byte[ methodBlockAdr + 0 ] >> jemFileCountAdjust
  numberOfLocals     := byte[ methodBlockAdr + 1 ] >> jemFileCountAdjust
  flags              := byte[ methodBlockAdr + 2 ]
  methodAdr          := byte[ methodBlockAdr + 3 ] << 8
  methodAdr          |= byte[ methodBlockAdr + 4 ]
  exceptionTable     := byte[ methodBlockAdr + 5 ] << 8
  exceptionTable     |= byte[ methodBlockAdr + 6 ]


  if sp > @stack
    oldFramePointer := fp
    fp := sp
  else
    oldFramePointer := 0
    fp := 0
    
  if callDebug
    tv.Str( string($d,"CallMethod "))
    tv.Hex( methodAdr, 4 )
    tv.Str( string(" Block@ "))
    tv.Hex( methodBlockAdr-bytecodeStart, 4 )
    tv.Str( string(" #P "))
    tv.Hex( numberOfParameters, 2 )
    tv.Str( string(" #L "))
    tv.Hex( numberOfLocals, 2 )
    tv.Str( string(" ET "))
    tv.Hex( exceptionTable, 4 )
    tv.Str( string(" Fl "))
    tv.Hex( flags, 2 )
    tv.Str( string(" oldFP "))
    tv.Hex( oldFramePointer, 4 )

  '               SP -->
  '       .----.        .----.          .----.          .----.
  '       | A  |        |    |          | SP |          | SP |
  ' SP -->| B  |        :    :          :    :          :    :
  '       `----'        |    |          | BP |          | BP |
  '                     :    :          :    :          :    :
  '                     |    |          | PC |          | PC |
  '                     |----|          |----|          |----|
  '                     | A  |    BP -->| A  |    BP -->| A  |
  '                     | B  |    SP -->| B  |          | B  |
  '                     `----'          `----'          |----|
  '                                                     | 0  |
  '                                                     | 0  |
  '                                               SP -->| 0  |
  '                                                     `----'
  '
  ' 1) Adjust SP to below parameters, shuffle parameters up
  ' 2) Insert stack frame, adjust BP and SP
  ' 3) Push zeroed local variables

  sp -= ( (numberOfParameters) << INTWORDS )

  ' For InvokeVirtual, we need to keep a record of the working class.
  ' The working class excludes java.lang.Object and init.
  ' During InvokeVirtual, the method to invoke is represented as
  ' an offset from the class address or classid.
  '
   
  gMethodId := methodBlockAdr-bytecodeStart
  gClassId := word[fp+1]
  
  if callDebug
    classid := LookupMethodClassId(methodBlockAdr-bytecodeStart)
    tv.Str( string(" gClassID $"))
    tv.Hex( classid, 4 )
    tv.Str( string(" MethodId $"))
    tv.Hex( gMethodId, 4 )

  if numberOfParameters <> 0

    frameStackSize := ( ( @sp - @pc ) >> 2 ) + 1 ' How many PushAddr's follow

    if not INT32
      if ChipVer <> 1           
        frameStackSize <<= 1
                                
    repeat i from 1 to numberOfParameters           
      Poke( sp, i + frameStackSize, Peek( sp, i ) )
  
  repeat i from @sp to @pc step 4 ' Not -4 ! That's a Propeller Gotcha
    PushAddr( long[ i ] )
                  
  bp := sp + INTBYTES

  sp += ( numberOfParameters << INTWORDS )

  repeat numberOfLocals
    Push( 0 )

  ' peter's frame info
  '
  Push(gMethodId)              
  Push(oldFramePointer)
  Push(pc-bytecodeStart)

  et := exceptionTable + bytecodeStart

  case flags
  
    $00 :   pc := methodAdr + bytecodeStart
            mb := pc
            if 0'hasClinit
              repeat while byte[ mb ] == jem#Jem_INVOKESTATIC
                mb += 3
              pc := mb ' ****** HACK **** This skips all "invokestatic <??.<clinit>>" at start of 'main'

    $80 :   if NativeMethod( methodAdr, numberOfParameters )
              ReturnFromMethodWithValue
            else
              ReturnFromMethod

    other :
      tv.out($d)
      tv.out($d)
      tv.hex(flags, 2)
      Crashed( String( "Unknown Method Table Flag value" ) )
  
PRI ReturnFromMethodWithValue

  '       .----.                  .----.
  '       | ?? |            BP -->| ?? |  <------.
  '       :    :                  :    :         |
  '       | ?? |                  | ?? |  <--.   |
  '       |----|                  |----|     |   |
  '       | SP |------.     SP -->| XX |     |   |
  '       :    :      |           `----'     |   |
  '       | BP |---.  |                      |   |
  '       :    :   |  `----------------------'   |
  '       | PC |   |                             |
  '       |----|   `-----------------------------'
  ' BP -->| A  |
  '       | B  |
  '       |----|
  ' SP -->| XX |
  '       `----'
  
  result := Pop
  ReturnFromMethod
  Push( result )
{
  if jazzDebug
    tv.Str( string($d," .. ReturnFromMethod with value "))
    tv.Hex(result,4)
}
      
PRI ReturnFromMethod | i, oldpc, oldfp, oldmd

  ' Note that it would not be essential to have pushed the SP in the current implementation as the SP
  ' is always the address preceding the stack frame and could therefore be determined as an offset from
  ' BP ( and also SP would point to where it should be when PC, MB and BP have been popped ) but this
  ' is designed to pre-empt stack frames having been allocated from the global heap at some time in the
  ' future.
  ' 
  '       .----.                  .----.
  '       | ?? |            BP -->| ?? |  <------.
  '       :    :                  :    :         |
  '       | ?? |            SP -->| ?? |  <--.   |
  '       |----|                  `----'     |   |
  '       | SP |-----------------------------'   |
  '       :    :                                 |
  '       | BP |---------------------------------'
  '       :    :
  '       | PC |
  '       |----|
  ' BP -->| A  |
  ' SP -->| B  |
  '       `----'

  ' pop peter's 
  oldpc := Pop 'from Push(pc)
  oldfp := Pop 'from Push(oldFramePointer)
  oldmd := Pop 'from Push(gMethodId)              
  
  sp := bp - INTBYTES

  repeat i from @pc to @sp step 4
    long[ i ] := PopAddr
{
  if jazzDebug
    tv.Str( string($d," .. ReturnFromMethod pc $"))
    tv.Hex(oldpc,4)
    tv.Str( string(" .. OldFP $"))
    tv.Hex(oldfp,4)
    tv.Str( string(" .. Method $"))
    tv.Hex(oldmd,4)
}
    
' .-----------------------------------------------------------------------------------------------------.
' |     Call Jvm PASM Handler                                                                           |
' `-----------------------------------------------------------------------------------------------------'

PRI JvmPasm

  result := false
  
' .-----------------------------------------------------------------------------------------------------.
' |     Utility Code                                                                                    |
' `-----------------------------------------------------------------------------------------------------'

PRI ShowStack | len, ii

  if stackAltered
    stackAltered := false
    tv.Out( $0D )
    tv.str(string("ST "))
    tv.Hex( sp, 4 )
    tv.Out( " " )
    len := (( sp - @stack ) >> INTWORDS) + 1
    tv.Dec( len)
    'tv.Hex( ( ( sp - @stack ) >> INTWORDS ) + 1, 2)
    tv.out( " " )
    tv.out( $d )
    'len := 15
    if  INT32
      repeat ii from 0 to len/2 
        if ii and (ii // 16) == 0
          tv.out($d)
        ShowStackLong( ii )
    else
      repeat ii from 0 to len 
        if ii and (ii // 16) == 0
          tv.out($d)
        ShowStackWord( ii )
  
PRI showStackLong( depth )

  tv.out( " " )
  if ( ( sp - @stack ) >> 2 ) => depth
    tv.Hex( long[ sp - (depth<<2) ], 8 )
  else
    tv.Str( String("--------") )

PRI showStackWord( depth )

  tv.out( " " )
  if ( ( sp - @stack ) >> 1 ) => depth
    tv.Hex( word[ sp - (depth<<1) ], 4 )
  else
    tv.Str( String("----") )

PRI ShowHeap | ptr, typ, bytLen, datPtr, nextPtr

  ptr := heap
  repeat while ptr < HEAP_BASE
    nextPtr := GetHeapInfo( ptr, @typ, @bytLen, 0 )
    ShowHeapEntry( ptr, typ, bytLen )
    ptr := nextPtr
    
PRI ShowHeapEntry( ptr, typ, bytLen )

  tv.Out( $0D )
  tv.Hex( ptr, 4 )
  tv.Out( "-" )
  if ( ChipVer == 1 ) or ( not BIGARRAYS )
    tv.Hex( ptr + bytLen + ( HEAP_HEAD_DATA << HEAP_WORD ) - 1, 4 )
  else
    tv.Hex( ptr + bytLen + ( HEAP_HEAD_DATA << HEAP_LONG ) - 1, 4 )
  tv.Out( " " )
  tv.Dec( typ )
  if typ < 10
    tv.Out( " " )
  tv.Out( " " )
  tv.Str( jem.TypeNamePtr(typ) )
  tv.Dec( BitSizePerType(typ) ) 
  tv.Out( "[" )
  tv.dec( bytLen >> BytesShiftPerType(typ) )
  tv.Out( "]" )  
  tv.Out( " " )
  if ( ChipVer == 1 ) or ( not BIGARRAYS )
    tv.Dec( HEAP_HEAD_DATA << HEAP_WORD )
  else
    tv.Dec( HEAP_HEAD_DATA << HEAP_LONG )
  tv.Out( "+" )
  tv.Dec( bytLen )
  tv.Str( String( " Bytes" ) )
  if typ & $80
    tv.Str( String( " DEAD" ) )
    
PRI BitSizePerType(typ)

  result := 8 << BytesShiftPerType(typ)

PRI CompactHeap | ptr, typ, bytLen, datPtr, nextPtr, nextTyp
{
  This should work but not tested and never called
  
  ptr := heap
  repeat while ptr < HEAP_BASE
    nextPtr := GetHeapInfo( ptr, @typ, @bytLen, @datPtr )
    if typ & jem#T_DELETED
      if ptr == heap
        heap := nextPtr
      else
        if nextPtr < HEAP_BASE
          GetHeapInfo( nextPtr, @nextTyp, 0, 0 )
          if nextTyp & jem#T_DELETED
            ' Combine the two
            nextPtr := GetHeapInfo( nextPtr, 0, 0, 0 )
            if ( ChipVer == 1 ) or ( not BIGARRAYS )
              word[ ptr + ( HEAP_HEAD_BLEN << HEAP_WORD ) ] := nextPtr - datPtr 
            else
              long[ ptr + ( HEAP_HEAD_BLEN << HEAP_LONG ) ] := nextPtr - datPtr  
            nextPtr := ptr
    ptr := nextPtr
}

               
PRI Crashed( txtPtr )

  dumpRegs
  ShowStack
  ShowHeap
  tv.Out( $0D )
  tv.Str( txtPtr )
  abort jvmCrashed
  repeat

PRI Pause_Ms( ms )

  waitcnt( CLKFREQ / 1000 * ms + CNT )


VAR ' gClassIdArray
  long gClassIdArray  ' heap pointer to class id array with id's stored by index in order found
  long gMethodIdArray ' heap pointer to array of method id arrays 
  
DAT ' ClassLoad method descriptor
  methNumParms          byte 0
  methNumLocals         byte 0
  methFlags             byte 0
  methOffset            word 0
  methExcTable          word 0
  methClass             word 0
  methAddr              word 0

CON
  methConsLen = 7               ' length of method constant 
  methDescLen = 4*2+3

PRI LoadClassInfo(mainMethodBase) | listIndex, superClass, classAddr, classFields, numClass, numMethod, methArray, data, type, len
{{
/**
 * Load a class info from the jem code binary file class table. This is an embedded
 * implementation so we don't look for files; everything is in the binary.
 * This table is used for classid lookup. ANEWARRAY is one code that needs this.
 *
 * The method builds two arrays: classid array and methodid array
 * The classid array is a list of class id's arranged by order found.
 * The methodid array is a list of arrays where each array is bound
 * to a class by classid index and contains method id's.
 *
 * @param mainMethodBase - this is pc address of the main method
 * @returns pointer to classlist heap "array"
 */
LoadClassInfo(mainMethodBase);
}}
  ' all classes have a super class except java.lang.Object and it's address is -1
  ' all classes have a fields field. all classes have at least one method.
{  
  if 0'jazzDebug
    tv.str(string($d,"LoadClassInfo $"))
    tv.hex(bytecodeStart+20,4)
    tv.str(string(" main addr $"))
    tv.hex(mainMethodBase,4)
}

  ' count classes
  '
  numClass := 0

  ' if we run into a new class, methExcTable will be non-zero.
  ' if we exceed class definitions, classAddr will be > mainMethodBase
  '
  ' in this loop we create stack entries. the end result will be
  ' classAddr0, methArray0, classAddr1, methArray1 ... classAddrN, methArrayN
  ' in the middle, methAddrs are pushed then popped to the methArrays
  ' 
  methExcTable := 0
  listIndex := bytecodeStart+20  ' class info starts at 20
  repeat while (listIndex < mainMethodBase)
    ' class address
    classAddr := listIndex-bytecodeStart
    Push(classAddr)
    superClass  := byte[listIndex++] << 8
    superClass  |= byte[listIndex++]
    classFields := byte[listIndex++]
{
    if 0'jazzDebug
      tv.str(string($d,"cAddr $"))
      tv.hex(classAddr,4)
      tv.str(string(" supC $"))
      tv.hex(superClass,4)
      tv.str(string(" cFlds "))
      tv.dec(classFields)
}     
    ' cound methods
    '
    numMethod := 0
    repeat while (listIndex < mainMethodBase) ' class info starts at 20
      bytemove(@methNumParms, listIndex, methConsLen)
      if methExcTable <> 0
        quit
      else
        methClass := classAddr
        methAddr := listIndex-bytecodeStart
        Push(methAddr)
        numMethod++
{
        if 0'jazzDebug
          tv.str(string($d,"// classSig $"))
          tv.hex(methClass,4)
          tv.str(string(" methAddr $"))
          tv.hex(methAddr,4)
          tv.str(string($d,"@methNumParms $"))
          tv.hex(@methNumParms,4)
          tv.str(string(" val $"))
          tv.hex(methNumParms,2)
          tv.str(string($d,"@methNumLocals $"))
          tv.hex(@methNumLocals, 4)
          tv.str(string(" val $"))
          tv.hex(methNumLocals, 2)
          tv.str(string($d,"@methFlags $"))
          tv.hex(@methFlags, 4)
          tv.str(string(" val $"))
          tv.hex(methFlags, 2)
          tv.str(string($d,"@methOffset $"))
          tv.hex(@methOffset,4)
          tv.str(string(" val $"))
          tv.hex(methOffset,4)
          tv.str(string($d,"@methExcTable $"))
          tv.hex(@methExcTable,4)
          tv.str(string(" val $"))
          tv.hex(methExcTable,4)
}
      listIndex += methConsLen

    methArray := GetFromHeap(jem#T_SHORT,numMethod) 
    GetHeapInfo(methArray, @type, @len, @data)

    repeat while(numMethod-- > 0)
      methAddr := Pop
      SetHeapDataValue(data, type, len, numMethod, methAddr)
{
      if 0'jazzDebug
          tv.str(string($d,"classNum "))
          tv.dec(numClass)
          tv.str(string(" classAddr $"))
          tv.hex(classAddr,4)
          tv.str(string(" methAddr $"))
          tv.hex(methAddr,4)
}
    Push(methArray)
{          
    if 0'jazzDebug
      tv.str(string($d,"cCount $"))
      tv.hex(numClass,4)
      tv.str(string(" mCount $"))
      tv.hex(numMethod,4)
}
    numClass++


  ' create storage for method ids.
  '
  gMethodIdArray := GetFromHeap(jem#T_SHORT,numClass)
{
  if jazzDebug 
    GetHeapInfo(gMethodIdArray, @type, @len, @data) ' debug only
}
  ' create storage for class ids
  '  
  result := GetFromHeap(jem#T_SHORT, numClass)
  GetHeapInfo(result, @type, @len, @data)

  ' put method arrays in class num indexed method array storage
  ' put classid = classAddr values in class list
  '  
  repeat while(numClass-- > 0)
    methArray := Pop
    SetHeapValue(gMethodIdArray, numClass, methArray)
    classAddr := Pop
    SetHeapDataValue(data, type, len, numClass, classAddr)
{
    if 0'jazzDebug
      tv.str(string($d,"classNum "))
      tv.dec(numClass)
      tv.str(string(" classAddr $"))
      tv.hex(classAddr,4)
}  
  GetHeapInfo(result, 0, 0, @data)
  gClassIdArray := result

  
PRI LookupClassIdIndex(classid) | ii, type, len, data, cid
{{
/**
 *
 * Get classid array index.
 * @param classid - classid is hex 4 signature at class bytecode address
 * @returns classid's 0..classnum or -1 on failure
 */
LookupClassIdIndex(classid);
}}
{
  if 0'jazzDebug
    tv.str(string($d,"// LookupClassIdIndex $"))
    tv.hex(classid,4)
}
  result := -1 

  GetHeapInfo(gClassIdArray, @type, @len, @data)
  repeat ii from 0 to (len-1)>>BytesShiftPerType(type)
    cid := GetHeapDataValue(data, type, len, ii)
    'cid := GetHeapValue(gClassIdArray, ii)  ' another way
    'cid := word[data+ii<<1]                 ' easiest way to get it wrong
{
    if 0'jazzDebug
      tv.str(string($d,"Lookup $"))
      tv.hex(classid,4)
      tv.str(string(" classId $"))
      tv.hex(cid,4)
}
    if cid == classid
{
      if 0'jazzDebug
        tv.str(string($d,"/* LookupClassIdIndex $"))
        tv.hex(classid,4)
        tv.str(string(" = "))
        tv.dec(ii)
        tv.str(string(" */ "))
}
      return ii

PRI LookupClassAtIndex(index) | ii, type, len, data, cid
{{
/**
 *
 * Get classid array index.
 * @param index - class index of classid
 * @returns classid at index or 0 on failure
 */
LookupClassAtIndex(index);
}}
{
  if 1'jazzDebug
    tv.str(string($d,"// LookupClassAtIndex "))
    tv.dec(index)
}
  result := 0 

  GetHeapInfo(gClassIdArray, @type, @len, @data)
  if index < len and index => 0
    result := GetHeapDataValue(data, type, len, index)

PRI LookupMethodIndex(classid, methid) | ii, type, len, data, methArray, cid
{{
/**
 *
 * Get method array index in a class
 *
 * @param classid - classid is hex 4 signature at class bytecode address
 * @param methid  - methid  is hex 4 signature at method bytecode address
 *
 * @returns classid's method num 0..methonum or -1 on failure
 */
LookupMethodIndex(classid, methid);
}}
{
  if 0'jazzDebug
    tv.str(string($d,"// LookupMethodIndex $"))
    tv.hex(classid,4)
    tv.str(string(" $"))
    tv.hex(methid,4)
}
  result := -1
   
  cid := LookupClassIdIndex(classid)
  if cid < 0
    return result
    
  GetHeapInfo(gMethodIdArray, @type, @len, @data)
  methArray := GetHeapDataValue(data, type, len, cid)
  if methArray == 0
    return result
  GetHeapInfo(methArray, @type, @len, @data)
  repeat ii from 0 to (len-1)>>BytesShiftPerType(type)
    result := GetHeapDataValue(data, type, len, ii)
{
    if 0'jazzDebug
      tv.str(string($d,"Lookup $"))
      tv.hex(classid,4)
      tv.str(string(" classId $"))
      tv.hex(result,4)
}      
    if result == methid
{
      if 0'jazzDebug
        tv.str(string($d,"/* LookupMethodIndex $"))
        tv.hex(classid,4)
        tv.str(string(" $"))
        tv.hex(methid,4)
        tv.str(string(" = $"))
        tv.hex(ii,4)
        tv.str(string(" */ "))
}
      return ii


PRI LookupMethodClassId(methid) | ii, jj, ctype, clen, cdata, mtype, mlen, mdata, methArray, type, len, data, cid
{{
/**
 * Get method's class id ... class id is bytecode offset of class definition.
 * @param methid  - methid  is hex 4 signature at method bytecode address
 * @returns classid's method num 0..methonum or -1 on failure
 */
LookupMethodClassId(methid);
}}
  result := 0
{
  if 0'jazzDebug
    tv.str(string($d,"// LookupMethodClassId $"))
    tv.hex(methid,4)
}   
  GetHeapInfo(gClassIdArray, @ctype, @clen, @cdata)
  GetHeapInfo(gMethodIdArray, @mtype, @mlen, @mdata)
  repeat ii from 0 to (clen-1)>>BytesShiftPerType(type)
    methArray := GetHeapDataValue(mdata, mtype, mlen, ii)
    GetHeapInfo(methArray, @type, @len, @data)
    repeat jj from 0 to (len-1)>>BytesShiftPerType(type)
      result := GetHeapDataValue(data, type, len, jj)
{
      if 0'jazzDebug
        tv.str(string($d,"Lookup $"))
        tv.hex(ii,4)
        tv.str(string(" methId $"))
        tv.hex(result,4)
}
      if result == methid
        result := GetHeapDataValue(cdata, ctype, clen, ii)
{
        if 0'jazzDebug
          tv.str(string($d,"LookupMethodClassId result $"))
          tv.hex(result,4)
}
        return result

  
CON

  #0, ObjectThisClass, ObjectSuperClass, ObjectFields, ObjectLastEnum ' field enum should be last

{
Objects have this pointer, superclass pointer, constant table pointer (array of arrays), and field pointers.
}
  
PRI doNEW | object
{{
/**
 * See case jem#Jem_NEW comment for basic operation description.
 * Creates a new class object and puts reference on stack.
 */
doNEW();
}}     
  'classAddr := Operand_BytecodeAddress
  object := GetNewClass(Operand_BytecodeAddress)
  Push(object)
  
  result := 0
  
PRI GetNewClass(classAddr) | object, ii, fieldPtr, data, superClass, numFields, type, len
{{
/**
 * 
 * Create a new class object with classAddr resulting in a reference which is a heap pointer
 * with entries populated according to the CON #0 enum above ... ObjectThisClass, .....
 * Constant table is not loaded in each object ... this could change if necessary ...
 * constant table is a global array of String objects for LDC to access.
 *
 * @returns object reference
 */
GetNewClass(classAddr);
}}     

{
  if jazzDebug
    tv.Str( String($d,"New " ) )
    tv.Hex( pc-bytecodeStart, 4 )
}  
  superClass := byte[classAddr+0] << 8
  superClass |= byte[classAddr+1]
  numFields  := byte[classAddr+2] >> 1
   
  ' allocate heap pointer for class
  ' get enough storage for superClass and numFields
  ' fields are stored in the stack ???
  '
  object := GetFromHeap(jem#T_ANY, ObjectLastEnum) 
  GetHeapInfo(object, @type, @len, @data)
{ 
  if jazzDebug
    tv.Str(string($d,"NewClass pc "))
    tv.Hex(pc-bytecodeStart, 4 )
    tv.Str(string(" this "))
    tv.Hex(object, 4 )
    tv.Str(string(" class addr "))
    tv.Hex(classAddr, 4 )
    tv.Str(string(" SC "))
    tv.Hex(superClass, 4 )
    tv.Str(string(" F$"))
    tv.Hex(numFields, 2 )
    tv.Str(string(" t$"))
    tv.Hex(type, 4 )
    tv.Str(string(" l$"))
    tv.Hex(len, 2 )
} 
  'save this class name
  classAddr -= bytecodeStart
  SetHeapDataValue(data, type, len, ObjectThisClass, classAddr)
  SetHeapDataValue(data, type, len, ObjectSuperClass, superClass)
  fieldPtr := GetFromHeap(jem#T_ANY,numFields+1)
  SetHeapDataValue(data, type, len, ObjectFields, fieldPtr)
  repeat ii from 0 to numFields
    SetHeapValue(fieldPtr, ii, GetFromHeap(jem#T_ANY, 1))

  ' handle constant table globally if possible
  'SetHeapDataValue(data, type, len, ObjectConTable, GetConstantTable(object)) 'GetFromHeap(jem#T_ANY,GetConstantTableSize+1))
    
  result := object
  if jazzDebug
    ShowObject(result)

PRI GetConstantTableSize | constAddr
  constAddr := byte[ct + 0] << 8
  constAddr |= byte[ct + 1]
  constAddr += bytecodeStart 
  result := (constAddr - ct) / 2

PRI InitConstantTable | ii, jj, type, len, data, constAddr, constLen, constOff, constCnt, constSize, classAddr, ptr, object
  if gConstTable <> 0
    return gConstTable
    
  constSize := 10
  constAddr := byte[ct + 0] << 8
  constAddr |= byte[ct + 1]
  constAddr += bytecodeStart 

  constCnt := (constAddr - ct) / 2
{
  if jazzDebug
    tv.Str(string($d,"InitCT "))
    tv.Dec(constCnt)
    tv.Str(string(" cadr "))
    tv.Dec(constAddr-bytecodeStart)
    tv.Out(" ")
    tv.Hex(constAddr-bytecodeStart, 4)
    tv.Str(string(" ct "))
    tv.Dec(ct-bytecodeStart)
    tv.Out(" ")
    tv.Hex(ct-bytecodeStart, 4)
}
  gConstTable := GetFromHeap(jem#T_ANY,constCnt)
  if gConstTable == 0
    Crashed(string("InitConstantTable can't find object's table pointer"))

  repeat ii from 0 to constCnt-1
    constLen := byte[constAddr + 2] << 8
    constLen |= byte[constAddr + 3]
    constOff := constAddr + constSize

    classAddr := byte[constAddr+0] << 8
    classAddr |= byte[constAddr+1]
{
    if jazzDebug
      tv.Str(string($d,"InitCT "))
      tv.Dec(constCnt)
      tv.Str(string(" index "))
      tv.Dec(ii)
      tv.Str(string(" classadr "))
      tv.Hex(classAddr, 4 )
      tv.Str(string(" cadr "))
      tv.Hex(constAddr-bytecodeStart, 4 )
      tv.Str(string(" clen "))
      tv.Hex(constLen, 4 )
      tv.Str(string(" coff "))
      tv.Hex(constOff-bytecodeStart, 4 )
      tv.Out(" ")
      tv.Hex(constOff, 4 )
      tv.Out(" ")  
}
    object := GetNewClass(classAddr+bytecodeStart)

    ptr := GetFromHeap(jem#T_BYTE, constLen+1)
    
    GetHeapInfo(ptr, @type, @len, @data)
    SetObjectField(object, 0, classAddr)
    SetObjectField(object, 1, len)
    SetObjectField(object, 2, ptr)
    repeat jj from 0 to constLen
      SetHeapDataValue(data, type, len, jj, byte[constOff+jj])
    
    ' put const in object's constant table ... this may be moved later
    SetHeapValue(gConstTable, ii, object)
    
    'ShowObject(object)
    constAddr += constSize + constLen
    
  result := gConstTable

PRI ShowObject(object) | ii, ptr, cptr, fptr, type, len, data, clen, flen, cdata, fdata
{{
/**
 * This displays all of objects contents if jazzDebug is enabled.
 * ObjectThisClass, ObjectSuperClass, ObjectFields ... from enum are displayed
 * @param object - object pointer
 */
ShowObject(object);
}}
  'if jazzDebug
  'jazzDebug := 0
  GetHeapInfo(object, @type, @len, @data)  
  tv.Str(string($d," ShowObject type $"))
  tv.Hex(type, 4 )
  ptr := GetHeapDataValue(data, type, len, ObjectThisClass)
  tv.Str(string(" this $"))
  tv.Hex(ptr, 4)
  ptr := GetHeapDataValue(data, type, len, ObjectSuperClass)
  tv.Str(string(" Super $"))
  tv.Hex(ptr, 4 )
{    
  cptr := gConstTable 'GetHeapDataValue(data, type, len, ObjectConTable)
  tv.Str(string(" ConTable $"))
  tv.Hex(cptr, 4 )
  GetHeapInfo(cptr, 0, @clen, @cdata)
  repeat ii from 0 to clen>>1-1
    tv.out(" ")
    tv.hex(GetHeapDataValue(cdata, type, clen, ii),4)
}
  fptr := GetHeapDataValue(data, type, len, ObjectFields)
  tv.Str(string(" Fields $"))
  tv.Hex(fptr, 4 )
  GetHeapInfo(fptr, 0, @flen, @fdata)
  repeat ii from 0 to flen>>1-1
    tv.out(" ")
    tv.hex(GetHeapDataValue(fdata, type, len, ii),4)
  'jazzDebug := 1
 
              
PRI GetObjectField(object, index) | ptr, type, len, data 
{{
/**
 * gets object var field[index] value
 * @param object - object pointer
 * @param index - index to objects field
 * @returns value to from field[index]
 */
GetObjectField(objectref, index);
}}
  if jazzDebug
    ShowObject(object)
  ptr := GetHeapValue(object,ObjectFields)
  if ptr == 0
    Crashed(string("GetObjectField can't find object's field pointer"))

  GetHeapInfo(ptr, @type, @len, @data)
  if index =< len  
    result := GetHeapDataValue(data, type, len, index)
  else
    tv.Str(string($d,"GetObjectField "))
    tv.Dec(index)
    tv.Str(string(" < len ? "))
    tv.Dec(len)
    Crashed(string(" GetObjectField index out of bounds"))
{
  if jazzDebug
    tv.Str(string($d,"GetObjectField "))
    tv.Dec(index)
    tv.Str(string(" value "))
    tv.Hex(result, 4 )
}
           
PRI SetObjectField(object, index, value) | ptr, type, len, data
{{
/**
 * sets object var field[index] = value
 * @param object - object pointer
 * @param index - index to objects field
 * @param value - value to store in field[index]
 */
SetObjectField(objectref, index, value);
}} 
  if jazzDebug
    ShowObject(object)
{
  if jazzDebug
    tv.Str(string($d,"SetObjectField "))
    tv.Dec(index)
    tv.Str(string(" value "))
    tv.Hex(value, 4 )
}
  ptr := GetHeapValue(object,ObjectFields)
  if ptr == 0
    Crashed(string("SetObjectField can't find object's field pointer"))
  GetHeapInfo(ptr, @type, @len, @data)
  if index =< len  
    SetHeapDataValue(data, type, len, index, value)
  else
    tv.Str(string($d,"SetObjectField "))
    tv.Dec(index)
    tv.Str(string(" < len ? "))
    tv.Dec(len)
    Crashed(string("SetObjectField index out of bounds"))
             
PRI LoadConstant(constantSizeInBytes, tableIndex) | ii, constAdr, constLen, constOff, classAddr, objectref, ptr, type, len, data
{{
/**
 * This loads a constant ....
 * @todo evaluate how int, long, and double will be treated etc....
 * @param constantSizeInBytes -  using one of two methods ....
 *  If constant size == 1, load the String object from global constant table at tableIndex.
 *  If constant size > 1, load the constant offset value. LIKELY BROKEN.
 * @param tableIndex -  index of constant to load.
 * @returns 0 to let case know the function is implemented
 */
LoadConstant(constantSizeInBytes, tableIndex);
}}

  if gConstTable == 0
    Crashed(string("GetObjectConstant can't find object's table pointer"))

  if constantSizeInBytes > 1
    constAdr := byte[ct + tableIndex << 1 + 0]
  else
    constAdr := byte[ct + tableIndex << 1 + 0] << 8
    constAdr |= byte[ct + tableIndex << 1 + 1]
     
  constAdr += bytecodeStart 

  constLen := byte[constAdr + 8] << 8
  constLen |= byte[constAdr + 9]
  constOff := constAdr + 10
{
  if jazzDebug
    tv.Str(string($d,"LDC "))
    tv.Dec(constantSizeInBytes)
    tv.Str(string(" "))
    tv.Hex(tableIndex, 4 )
    tv.Str(string(" cadr "))
    tv.Hex(constAdr-bytecodeStart, 4 )
    tv.Str(string(" clen "))
    tv.Hex(constLen, 4 )
    tv.Str(string(" coff "))
    tv.Hex(constOff-bytecodeStart, 4 )
    tv.Out(" ")
    tv.Hex(constOff, 4 )
    tv.Out(" ")  
}
  ' ldc2_w can only be a long/double
  '
  if constantSizeInBytes == 4
    Push(word[constOff])
    Push(word[constOff+1])
    return 0

  Push(GetHeapValue(gConstTable, tableIndex)) 
  result := 0  


PRI doGETFIELD | object, index, data
{{
/**
 * case jem#Jem_GETFIELD Get 16bit field from object.
 * Pop object/Push value
 * @returns 0 to let case know the function is implemented
 */
doGETFIELD();
}}
  index := Operand_UnsignedWord >> 1

  object := Pop
{
  if jazzDebug
    tv.Str(string($d,"doGETFIELD "))
    tv.Hex(object, 4)
    tv.Str(string(" index "))
    tv.Dec(index)
}
  data := GetObjectField(object, index)
{
  if jazzDebug
    tv.Str(string(" doGETFIELD value "))
    tv.Hex(data, 4 )
}
  Push(data)

  result := 0

    
PRI doPUTFIELD | object, index, value
{{
/**
 * case jem#Jem_PUTFIELD Set 16bit field in object.
 * Pop value/Pop object
 * @returns 0 to let case know the function is implemented
 */
doPUTFIELD();
}}
  index := Operand_UnSignedWord >> 1
  
  value := Pop
  object := Pop
{
  if jazzDebug
    tv.Str(string($d,"doPUTFIELD "))
    tv.Hex(object, 4)
    tv.Str(string(" index "))
    tv.Dec(index)
    tv.Str(string(" value "))
    tv.Hex(value, 4 )
}
  SetObjectField(object, index, value)
    
  result := 0
  
PRI doNEWARRAY | arrayLen, arrayType
{{
/**
 *
 * case jem#Jem_NEWARRAY Create new array of operand type
 * NEWARRAY                                                                        
 * Meaning: Allocate new array.                                                    
 * Description: A new array of a specific array type, capable of holding size      
 * elements, is allocated. The result is a reference to the new object. Allocation 
 * of an array large enough to contain size items of the specific array type is    
 * attempted and all elements of the array are initialized to 0.                   
 * size represents the number of elements in the new array and must be an          
 * integer. The result is stored with an internal code that indicates the type of  
 * array to allocate. Possible values for the type of array are as follows:        
 * T_BOOLEAN(4), T_chAR(5), T_FLOAT(6), T_DOUBLE(7), T_BYTE(8), T_SHORT(9),        
 * T_INT(10), and T_LONG(11).                                                      
 * Note                                                                            
 * A NegativeArraySizeException is thrown if size is less than 0. An               
 * OutOfMemoryError is thrown if there is not enough memory to allocate the array. 
 *
 * @returns 0 to let case know the function is implemented
 */
doNEWARRAY();
}}
  arrayLen := Pop
  'Operand_UnsignedByte ' ***** HACK **** This is $00 so what's it mean ?
  arrayType := Operand_UnsignedByte  ' $00 means *ANY* arrayType
  
  case arrayType
    0: arrayType := jem#T_INT
    1: arrayType := jem#T_BYTE
    other: arrayType := jem#T_ANY
{    
  if jazzDebug
    tv.Str(string($d,"NEWARRAY "))
    tv.Dec(arrayType)
    tv.Str(string(" "))
    tv.Hex(arrayLen, 4 )
    tv.Out(" ")  
}  
  Push( NewArrayRef( arrayType, arrayLen+1 ) )

  result := 0
  'debugMsDelay := -1

   
PRI doANEWARRAY | arrayLen, index, ptr, ii
{{
/**
 * case jem#Jem_NEWARRAY Create new array of operand type
 * ANEWARRAY                                                                      
 * Meaning: Allocate new array of objects.                                        
 * Description: A new array of the indicated class type and capable of holding    
 * size elements is allocated. The result is a reference to the new object.       
 * Allocation of an array large enough to contain size elements of the given class
 * type is attempted and all elements of the array are initialized to null.       
 * size represents the number of elements in the new array and must be an         
 * integer. byte1 and byte2 are used to construct an index into the constant pool 
 * of the current class. When the item at that index is resolved, the resulting   
 * entry must be a class.                                                         
 * The anewarray instruction is used to create a single-dimension array.          
 * Note                                                                           
 * A NegativeArraySizeException is thrown if size is less than 0. An              
 * OutOfMemoryError is thrown if there is not enough memory to allocate the array.
 * @returns 0 to let case know the function is implemented
 */                                                            
doANEWARRAY();
}}
  arrayLen := Pop

  ' index into the class constant pool ... not sure what it means for now.
  index := Operand_UnsignedWord
   
  ptr := NewArrayRef(jem#T_ANY, arrayLen)

  ' set all entries to be null
  repeat ii from 0 to arrayLen-1
    SetHeapValue(ptr, ii, 0)
    
  Push(ptr)
  
  result := 0

PRI doINVOKEVIRTUAL | methodAddr, methOff, method, ii, classid, ptr, type, len, data, offset, oldclass
{{
/**
 * Will call virtual method when developed.
 * @TODO: Make this work :)
 * @returns non-zero for now because it is not compeletely implemented.
 */
doINVOKEVIRTUAL();   
}}

  ' garbage for now
  offset := Operand_UnSignedByte >> 1
  offset--
  methOff := Operand_UnsignedByte
{
  if jazzDebug
    tv.Str(string($d,"InvokeVirtual ... moff $"))
    tv.Hex(methOff, 4 )
    tv.str(string(" off "))  
    tv.Hex(offset, 4 )
}
  'ShowStack
  'ShowObject(Peek(sp,-offset))

  ' this gClassId business can most likely be replaced by
  ' getting same info from object at top of the stack.
  ' if the object is not at tos, invokevirtual will fail anyway.
  '
  'classid := LookupMethodClassId(gMethodId)
  classid := GetHeapValue(Peek(sp,-offset), ObjectThisClass)
{ 
  if jazzDebug
    tv.Str(string($d,"InvokeVirtual ... moff $"))
    tv.Hex(methOff, 4 )
    tv.str(string(" off "))  
    tv.Hex(offset, 4 )
    tv.str(string(" cid "))  
    tv.Hex(classid, 4 )
}
  methOff+=bytecodeStart
  methodAddr := methOff+classid
  method := byte[methodAddr + 3] << 8
  method |= byte[methodAddr + 4]            
{
  if jazzDebug
    tv.str(string(" madd "))  
    tv.Hex(methodAddr-bytecodeStart, 4 )
    tv.str(string(" method "))  
    tv.Hex(method, 4 )
}      
  'check if methodAddr points to a method in our class
  'methodAddr := Operand_UnSignedWord
  'classid := LookupMethodClassId(methodAddr)

  if classid == $118
    return 0
    
  if classid
    CallMethod(methodAddr, false)
    Push(oldclass)
    return 0
  else
    Crashed(string("InvokeVirtual method not found."))

  result := 0


PRI doINSTANCEOF | objectref, index, classid, type, len, data
{{
/**
 * find out if objectref is an instance of a class
 * INSTANCEOF                                                                      
 * Meaning: Determine if an object is of given type and return result.             
 * Description: Determines whether the referenced object can be cast to be a       
 * reference to an object of the expected class. This instruction will overwrite   
 * the object reference with 1 if the object is an instance of the expected class  
 * or one of its superclasses. Otherwise, the object reference is overwritten by 0.
 * @returns 0 to let case know the function is implemented
 */
doINSTANCEOF();
}}
  index := Operand_UnSignedWord  
  objectref := Pop

  GetHeapInfo(objectref, @type, @len, @data)
  classid := GetHeapDataValue(data, type, len, ObjectThisClass)
  
  if index == classid
    Push(1)
  else
    Push(0)
     
  result := 0


PRI ShowHeapPtr( heapPtr ) | ii, type, len, data

  if GetHeapInfo(heapPtr, @type, @len, @data) == 0
    return

  if jazzDebug 
    tv.str(string($d,"heapPtr "))
    tv.hex(heapPtr,4)
    tv.str(string(" type "))
    tv.hex(type, 4)
    tv.str(string(" len "))
    tv.hex(len, 4)
     
  if jazzDebug
    tv.str( string($d,"data"))
  repeat ii from 0 to len-1 step 2
    if jazzDebug
      tv.out(" ")
      tv.hex(word[data+ii],4)

PRI doJazzLib : rc | ii, heapPtr, heapData, blockAddr, tmpAddr
{{
/**
 * Function is called from "notimplemented" function
 * @returns run status 0, of ok other wise no-zero
 */
doJazzLib();
}}
  result := -1

  case opcode
     
    jem#Jem_NEW:
      result := doNEW

    jem#Jem_GETFIELD:
      result := doGETFIELD

    jem#Jem_PUTFIELD:
      result := doPUTFIELD

    jem#Jem_NEWARRAY:
      result := doNEWARRAY

    jem#Jem_ANEWARRAY:
      result := doANEWARRAY

    jem#Jem_INVOKEVIRTUAL:
      result := doInvokeVirtual

    jem#Jem_INSTANCEOF:
      result := doINSTANCEOF


CON
  MAX_NUM_VPS   = 6  'as defined in CPU.java
  nmPeriph0Bank = $80
  nmPeriph1Bank = nmPeriph0Bank + $10
  nmPeriph2Bank = nmPeriph0Bank + $20
  nmPeriph3Bank = nmPeriph0Bank + $30
  nmPeriph4Bank = nmPeriph0Bank + $40
  nmPeriph5Bank = nmPeriph0Bank + $50

DAT
  vpRambank     byte 0[MAX_NUM_VPS*16]   'registers for Virtual Peripherals (16 bytes per VP)
{
  frame         word 0   'reference frame 
  optop         word 0   'operand top 
  pc            word 0   'program counter 
  vars          word 0   'variable register 
}
  carry         word 0   'true if overflow/underflow after math operation

{{

/**
 * native functions
}}

PRI do_native(nm_function, nm_numparms): value | parm
  'parameters for native functions need to be retrieved
  parm := 0
  case nm_function
     0: nm_core_CPU_writeRegister(GetVar(parm++),GetVar(parm++))
     1: return nm_core_CPU_readRegister(GetVar(parm++))
     2: nm_core_CPU_writeObject(GetVar(parm++),GetVar(parm++))
     3: nm_core_CPU_delay(GetVar(parm++))
     4: nm_core_CPU_message(GetHeapValuePtr(GetVar(parm++),0),GetVar(parm++))
     5: nm_page_
     6: return nm_core_CPU_count(GetVar(parm++),GetVar(parm++),GetVar(parm++))
     7: return nm_core_CPU_rcTime(GetVar(parm++),GetVar(parm++),GetVar(parm++))
     8: return nm_core_CPU_pulseIn(GetVar(parm++),GetVar(parm++),GetVar(parm++))
     9: nm_core_CPU_pulseOut(GetVar(parm++),GetVar(parm++))
    10: return nm_core_CPU_readPin(GetVar(parm++))
    11: nm_core_CPU_writePin(GetVar(parm++),GetVar(parm++))
    12: return nm_core_CPU_shiftIn(GetVar(parm++),GetVar(parm++),GetVar(parm++),GetVar(parm++))
    13: nm_core_CPU_shiftOut(GetVar(parm++),GetVar(parm++),GetVar(parm++),GetVar(parm++),GetVar(parm++))
    14: return nm_core_CPU_carry
    15: nm_page_
    16: nm_core_CPU_uninstall(GetVar(parm++))
    17: nm_core_CPU_install(GetVar(parm++))
    18: nm_core_CPU_nap(GetVar(parm++))
    19: nm_core_CPU_setOutput(GetVar(parm++))
    20: nm_page_
    21: nm_core_CPU_setInput(GetVar(parm++))
    22: nm_page_
    23: return nm_core_CPU_readSRAM(GetVar(parm++))
    24: nm_page_
    25: nm_core_CPU_writeSRAM(GetVar(parm++),GetVar(parm++))
    26: nm_page_
    27: return nm_core_EEPROM_eeRead(GetVar(parm++))
    28: nm_page_
    29: nm_core_EEPROM_eeWrite(GetVar(parm++),GetVar(parm++))
    30: nm_page_
    31: nm_core_Uart_txInit(GetVar(parm++))
    32: nm_core_Uart_rxInit(GetVar(parm++))
    33: nm_core_Uart_rxRead(GetVar(parm++))
    34: nm_core_Callback_returnTwo
    35: nm_core_Timer_init(GetVar(parm++))
    36: nm_core_Timer_latch(GetVar(parm++))
    37: nm_page_
    38: nm_core_PWM_install(GetVar(parm++))
    39: nm_page_
    40: nm_core_PWM_updateInternal(GetVar(parm++),GetVar(parm++),GetVar(parm++),GetVar(parm++),GetVar(parm++))
    41: nm_page_
    42: nm_core_DAC_install(GetVar(parm++))
    43: nm_page_
    44: nm_core_ADC_install(GetVar(parm++))
    45: nm_page_
    46: return nm_core_Terminal_getByte
    47: nm_page_


CON
  'port addresses used with readRegister and writeRegister
  sxRB = $06
  sxRC = $07

  PORTA = $00FF  'A constant representing the first I/O port on the Prop. This port contains pins 0-7.
  PORTB = $01FF  'A constant representing the second I/O port on the Prop. This port contains pins 8-15.


PRI nm_core_CPU_writeRegister(address,value)
  case address
    nmPeriph0Bank..(nmPeriph0Bank+MAX_NUM_VPS*16-1):
      vpRambank[address-nmPeriph0Bank] := value
    sxRB: outa[7..0] := value
    sxRC: outa[15..8] := value
    
  
PRI nm_core_CPU_readRegister(address):value
  case address
    nmPeriph0Bank..(nmPeriph0Bank+MAX_NUM_VPS*16-1):
      return vpRambank[address-nmPeriph0Bank]
    sxRB: return ina & $FF
    sxRC: return (ina >> 8) & $FF

  
PRI nm_core_CPU_writeObject(address,value)
  'Needs to write where ???


PRI nm_core_CPU_delay(period) | t
  '/**
  ' * Delays the program execution for a period of time. The delay method does not
  ' * return until the delay period has expired. During a delay any installed Virtual
  ' * Peripherals will continue to run.
  ' * <p>
  ' *
  ' * @param period the length of time to delay for, measured in
  ' * 95.48 microsecond units.
  ' */
  t := cnt
  repeat until (cnt - t) / (clkfreq * 95.48 / 1000000) > period

  
PRI nm_core_CPU_message(data,length)
  'write character array to debug output
  if length > 0
    repeat while length > 1
      'debug.cprintf(string("%c"),byte[data],true)
      tv.tx(byte[data])
      data++
      length--

PRI nm_core_CPU_count(timeout,portPin,edge): value | pin,t
  '/**
  ' * Count the number of rising or falling edge transitions that occur during
  ' * a set period of time.
  ' * If the pin is an output then it will be changed to be an input and will remain as
  ' * an input when the counting is complete.
  ' * <p>
  ' * To ensure reliable detection the input signal must maintain a value for at least
  ' * 8.68us. This means that a square wave must have a period of at least 17.36us and
  ' * a frequency less than 57.6kHz.
  ' *
  ' * @param timeout the amount of time to count transitions over. Measured in
  ' *                8.68us units. The value of 32,767 allows a maximum measurement
  ' *                time of 284.4ms.
  ' * @param portPin     the I/O pin to measure transitions on.
  ' * @param edge    the type of edge to count. Pass true to count rising edges,
  ' *                or false to count falling edges.
  ' * @return the number of rising edges or falling edges that occured on the pin
  ' *         during the timeout period.
  ' */
  {{ adapted from object BS2_functions PUB COUNT (Pin, Duration) : Value }}
  {{ only does rising-edge pulses now  }}
  if portPin < 256
    pin := >|portPin
  else
    portPin &= $FF
    pin := 8 + >|portPin
  dira[pin]~                                          ' Set as input
  ctra := 0                                           ' Clear any value in ctrb
                                                           ' set up counter, pos edge trigger
  ctra := (%01010 << 26 ) | (%001 << 23) | (0 << 9) | (pin)  
  frqa := 1                                           ' 1 count/trigger
  phsa := 0                                           ' Clear phase - holds accumulated count
  t := cnt
  repeat until (cnt - t) / (clkfreq * 8.68 / 1000000) > timeout ' Allow to count for duration in 8.68us units 
  value := phsa                                       ' Return total count
  ctra := 0                                           ' clear counter mode


PRI nm_core_CPU_rcTime(timeout,portPin,pinState): value | pin, ClkStart,clkStop
  '/**
  ' * Measure the length of time it takes for a pin to match a desired value.
  ' * If the pin is an output then it will be changed to be an input and will
  ' * remain as an input when rcTime is complete. rcTime can be used to measure
  ' * the charge/discharge time of a RC (resistor-capacitor) circuit in order to
  ' * read an analogue value.
  ' *
  ' * @param timeout the maximum measurable time.
  ' * @param portPin the I/O pin to measure.
  ' * @param pinState the pin state to wait for.
  ' * @return the amount of time taken for the pin to change to the desired value,
  ' * measured in 8.68us units. A value of -1 indicates that the timeout expired
  ' * before the desired state was reached.
  ' */
  {{ adapted from object BS2_functions PUB RCTIME (Pin,State):Duration | ClkStart, ClkStop }}
  if portPin < 256
    pin := >|portPin
  else
    portPin &= $FF
    pin := 8 + >|portPin
  if pinState
    pinState := 1
  else
    pinState := 0
  dira[pin]~
  ClkStart := cnt                                         ' Save counter for start time
  waitpne(pinState << pin, |< pin, 0)                     ' Wait for opposite state to end
  clkStop := cnt                                          ' Save stop time
  value := (clkStop - ClkStart)/(clkfreq * 8.68 /1000000) ' calculate in 8.68us resolution


PRI nm_core_CPU_pulseIn(timeout,portPin,pinState): value | pin
  '/**
  ' * Measure the length of a single pulse. Waits for the value of the selected
  ' * pin to match the value selected by the state parameter and then times how
  ' * long it takes for the pin to return to its original state.
  ' * If the pin is an output then it will be changed to be an input and will
  ' * remain as an input when complete.
  ' * <p>
  ' * A pulse of less than 8.68us in duration may be missed.
  ' *
  ' * @param timeout  the maximum amount of time to spend in the pulseIn method.
  ' *                 The timeout covers the complete time of the process both
  ' *                 the wait for the start of the pulse and the time of the pulse.
  ' * @param portPin      the I/O pin to measure the pulse on.
  ' * @param pinState    whether to measure a high pulse (true) or a low pulse (false).
  ' * @return the length of the pulse in 8.68us units. A value of 0 indicates
  ' *         that no start edge was detected. A value of -1 indicates that no
  ' *         stop edge was detected.
  ' */
  {{ adapted from object BS2_functions PUB PULSIN_Clk(Pin, State) : Duration }}
  {{ need to implement timeout }}
  if portPin < 256
    pin := >|portPin
  else
    portPin &= $FF
    pin := 8 + >|portPin
  dira[pin]~
  ctra := 0
  if pinState
    pinState := 1
  else
    pinState := 0
  if pinState == 1
    ctra := (%11010 << 26 ) | (%001 << 23) | (0 << 9) | (pin) ' set up counter, A level count
  else
    ctra := (%10101 << 26 ) | (%001 << 23) | (0 << 9) | (pin) ' set up counter, !A level count
  frqa := 1
  waitpne(pinState << pin, |< Pin, 0)                         ' Wait for opposite state ready
  phsa:=0                                                  ' Clear count
  waitpeq(pinState << pin, |< Pin, 0)                         ' wait for pulse
  waitpne(pinState << pin, |< Pin, 0)                         ' Wait for pulse to end
  value := phsa * 1000000 /clkfreq / 8.68                  ' Return duration as 8.68us units
  ctra :=0                                                 ' stop counter


CON
  cntMin     = 400      ' Minimum waitcnt value to prevent lock-up


PRI nm_core_CPU_pulseOut(length,portPin) | ClkCycles,pin
  '/**
  ' * Generate a pulse of a specific duration. If the pin is an input then it will
  ' * be changed to be an output and remain as an output when complete. To
  ' * ensure that a positive going pulse or negative going pulse is generated it
  ' * is a good idea to write a value to the pin first to put it into the correct
  ' * state. pulseOut inverts the value of the selected pin, waits for the
  ' * specified length of time, and then returns the pin to its original value.
  ' * <p>
  ' * The time is measured in 8.68us units with a maximum value of 32767 giving a
  ' * maximum pulse time of 284.4ms. A time of 0 will not cause a pulse.
  ' * pulseOut can be used to generate the control pulses for a servo motor.
  ' *
  ' * @param length the length of the pulse to generate, measured in 8.68us units.
  ' * @param portPin    the I/O to generate the pulse on.
  ' */
  {{ adapted from object BS2_functions PUB PULSOUT(Pin,Duration)  | clkcycles }}
  if portPin < 256
    pin := >|portPin
  else
    portPin &= $FF
    pin := 8 + >|portPin
  ClkCycles := (length * clkfreq / 1000000 * 8.68 - 1250) #> cntMin        ' duration * clk cycles for 8.68us
                                                           ' - inst. time, min cntMin
  dira[pin]~~                                              ' Set to output                                         
  !outa[pin]                                               ' set to opposite state
  waitcnt(clkcycles + cnt)                                 ' wait until clk gets there 
  !outa[pin]                                               ' return to orig. state


CON
  'pins
  pin0  = 1<<0
  pin1  = 1<<1
  pin2  = 1<<2
  pin3  = 1<<3
  pin4  = 1<<4
  pin5  = 1<<5
  pin6  = 1<<6
  pin7  = 1<<7
  pin8  = (1<<0)+256
  pin9  = (1<<1)+256
  pin10 = (1<<2)+256
  pin11 = (1<<3)+256
  pin12 = (1<<4)+256
  pin13 = (1<<5)+256
  pin14 = (1<<6)+256
  pin15 = (1<<7)+256
  
  
PRI nm_core_CPU_readPin(portPin): YesNo
  '/**
  ' * Reads the logic value on a pin. If the pin is an output then it will be
  ' * changed to be an input and will remain as an input when complete.
  ' *
  ' * @param portPin the I/O pin to read.
  ' * @return the state of the pin where true indicates high and false indicates low.
  ' */
  if portPin < 256
    dira[7..0] &= (255-portPin)
    return ina[>|portPin]
  else
    portPin -=256
    dira[15..8] &= (255-portPin)
    return ina[8 + >|portPin]
    

PRI nm_core_CPU_writePin(portPin,value)
  '/**
  ' * Set a pin to a logic state. Makes the selected pin high or low as selected
  ' * by the state parameter. If the pin is an input then it will be changed to
  ' * be an output and remain as an output when complete.
  ' *
  ' * @param portPin the I/O pin to write.
  ' * @param value the state to set the pin to (true=high, false=low).
  ' */
  if portPin < 256
    dira[7..0] |= portPin
    outa[>|portPin] := value
  else
    portPin -=256
    dira[15..8] |= portPin
    outa[8 + >|portPin] := value


CON
  'shiftin modes
  PRE_CLOCK_MSB  = 0  'Specify pre-clock sampling and MSB-first transmission. 
  POST_CLOCK_MSB = 1  'Specify post-clock sampling and MSB-first transmission.
  PRE_CLOCK_LSB  = 2  'Specify pre-clock sampling and LSB-first transmission.
  POST_CLOCK_LSB = 3  'Specify post-clock sampling and LSB-first transmission.
  OnClock        = 4

PRI nm_core_CPU_shiftIn(dataPortPin,clockPortPin,bitCount,mmode): value | Dpin,Cpin,InBit
  '/**
  ' * Read data from a synchronous serial device.
  ' * The clock pin outputs a clock signal and the data is clocked in form the data pin.
  ' * If the data pin is an output then it will be changed to be an input and remain as
  ' * an input when complete. If the clock pin is an input then it will be changed to be
  ' * an output and remain as an output when complete.
  ' * <p> The mode parameter selects pre/post clock sampling. Note that some devices
  ' * require pre-clock sampling. When pre-clock sampling is used the data pin is
  ' * sampled to obtain the first bit prior to any change of the clock pin. Clearly
  ' * the device must have had a clock pulse to output the first bit and this is
  ' * normally supplied by the previous call to shiftIn or shiftOut. Most
  ' * synchronous serial devices require a write (shiftOut) prior to reading and so the
  ' * clock for the first bit will be supplied by the shiftOut call. If the device must be
  ' * read without first being written to then the first read (shiftIn) should request an
  ' * extra bit to provide the initial clock pulse.
  ' * <p>
  ' * Note that the bit count can be set to higher than 16 (up to 256) if desired. When
  ' * a number greater than 16 is used the last 16 bits will be returned.
  ' *
  ' * @param dataPortPin the I/O pin to read data on.
  ' * @param clockPortPin the I/O pin on which to provide the clock to the external device.
  ' * @param bitCount the number of bits to count in (from 1 to 16). bitCount must not be 0.
  ' * @param mode the shifting and clocking mode to employ. This determines whether
  ' * bits are shifted into the least-significant-bit (LSb) or most-significant-bit (MSb)
  ' * first. The clocking mode determines whether the data is sampled before or after
  ' * the first clock transition. See the description for the possible values.
  ' * @return the bits shifted in on the data pin. The result contains bitCount bits and is
  ' * justified according to the mode setting. If MSb first is selected then the result
  ' * will be the data read in on the port. If LSb first is selected then the data is
  ' * returned MSb justified and must be shifted right (16-bitCount) bits to
  ' * obtain the correct value.
  ' */
  {{ adapted from object BS2_functions PUB SHIFTIN (Dpin, Cpin, Mode, Bits) : Value | InBit }}
{{
   Shift data in, master clock, for mode use BS2#MSBPRE, #MSBPOST, #LSBPRE, #LSBPOST
   Clock rate is ~16Kbps.  Use at 80MHz only is recommended.
     X := BS2.SHIFTIN(5,6,BS2#MSBPOST,8)
}}
  if dataPortPin < 256
    Dpin := >|dataPortPin
  else
    dataPortPin &= $FF
    Dpin := 8 + >|dataPortPin
  if clockPortPin < 256
    Cpin := >|clockPortPin
  else
    clockPortPin &= $FF
    Cpin := 8 + >|clockPortPin
    dira[Dpin]~                                            ' Set data pin to input
    outa[Cpin]:=0                                          ' Set clock low 
    dira[Cpin]~~                                           ' Set clock pin to output 
                                                
  If Mode == PRE_CLOCK_MSB                               ' Mode - MSB, before clock
       Value:=0
       REPEAT Bitcount                                         ' for number of bits
          InBit:= ina[Dpin]                                ' get bit value
          Value := (Value << 1) + InBit                    ' Add to  value shifted by position
          !outa[Cpin]                                      ' cycle clock
          !outa[Cpin]
          waitcnt(1000 + cnt)                              ' time delay

  elseif Mode == POST_CLOCK_MSB                            ' Mode - MSB, after clock              
       Value:=0                                                          
       REPEAT Bitcount                                         ' for number of bits                    
          !outa[Cpin]                                      ' cycle clock                         
          !outa[Cpin]                                         
          InBit:= ina[Dpin]                                ' get bit value                          
          Value := (Value << 1) + InBit                    ' Add to  value shifted by position                                         
          waitcnt(1000 + cnt)                              ' time delay                            
                                                                 
  elseif Mode == POST_CLOCK_LSB                            ' Mode - LSB, after clock                    
       Value:=0                                                                                         
       REPEAT Bitcount                                         ' for number of bits                         
          !outa[Cpin]                                      ' cycle clock                          
          !outa[Cpin]                                                                             
          InBit:= ina[Dpin]                                ' get bit value                        
          Value := (InBit << (bitcount-1)) + (Value >> 1)      ' Add to  value shifted by position    
          waitcnt(1000 + cnt)                              ' time delay                           

  elseif Mode == PRE_CLOCK_LSB                             ' Mode - LSB, before clock             
       Value:=0                                                                                   
       REPEAT Bitcount                                         ' for number of bits                   
          InBit:= ina[Dpin]                                ' get bit value                        
          Value := (Value >> 1) + (InBit << (bitcount-1))      ' Add to  value shifted by position    
          !outa[Cpin]                                      ' cycle clock                          
          !outa[Cpin]                                                                             
          waitcnt(1000 + cnt)                              ' time delay

  elseif Mode == OnClock  {{ is this required ?? }}                                          
       Value:=0
       REPEAT Bitcount                                         ' for number of bits
                                        
          !outa[Cpin]                                      ' cycle clock
          waitcnt(500 + cnt)                               ' get bit value
          InBit:= ina[Dpin]                               ' time delay
          Value := (Value << 1) + InBit                    ' Add to  value shifted by position
          !outa[Cpin]
          waitcnt(500 + cnt)                           
     



CON
  'shiftout modes
  SHIFT_MSB = 0  'Specify MSB-first transmission.
  SHIFT_LSB = 2  'Specify LSB-first transmission. 


PRI nm_core_CPU_shiftOut(dataPortPin,clockPortPin,bitCount,mmode,data) | Dpin,Cpin
  '/**
  ' * Write data to a synchronous serial device.
  ' * The clock pin outputs a clock signal and the data is clocked out the data pin. If
  ' * the data pin is an input then it will be changed to be an output and remain as an
  ' * output when complete. Likewise if the clock pin is an input then it will be
  ' * changed to be an output and remain as an output when complete.
  ' * <p>
  ' * Both the clock and data pins should be initialized before calling shiftOut. shiftOut
  ' * works by inverting the pins so you can generate either idle high or idle low signals.
  ' * <p>
  ' * If LSb first is selected then the data can be passed directly. If MSb first is
  ' * selected then if less than 16 bits are to be shift out, data should be shifted left
  ' * to MSb justify it.
  ' * @param dataPortPin the I/O pin to write data on.
  ' * @param clockPortPin the I/O pin on which to provide the clock to the external device.
  ' * @param bitCount the number of bits to count out (from 1 to 16). bitCount must not be 0.
  ' * @param mode the shifting mode to employ. This determines whether
  ' * bits are shifted into the least-significant-bit (LSb) or most-significant-bit (MSb)
  ' * first. Valid values are CPU.SHIFT_MSB and CPU.SHIFT_LSB.
  ' */
  {{ adapted from object BS2_functions PUB SHIFTOUT (Dpin, Cpin, Value, Mode, Bits)| bitNum  }}
{{
   Shift data out, master clock, for mode use ObjName#LSBFIRST, #MSBFIRST
   Clock rate is ~16Kbps.  Use at 80MHz only is recommended.
     BS2.SHIFTOUT(5,6,"B",BS2#LSBFIRST,8)
}}
  if dataPortPin < 256
    Dpin := >|dataPortPin
  else
    dataPortPin &= $FF
    Dpin := 8 + >|dataPortPin
  if clockPortPin < 256
    Cpin := >|clockPortPin
  else
  clockPortPin &= $FF
  Cpin := 8 + >|clockPortPin
  outa[Dpin]:=0                                          ' Data pin = 0
  dira[Dpin]~~                                           ' Set data as output
  outa[Cpin]:=0
  dira[Cpin]~~

  If Mode == SHIFT_LSB                                    ' Send LSB first    
       REPEAT Bitcount
          outa[Dpin] := data                              ' Set output
          data := data >> 1                              ' Shift value right
          !outa[Cpin]                                      ' cycle clock
          !outa[Cpin]
          waitcnt(1000 + cnt)                              ' delay

  elseIf Mode == SHIFT_MSB                                ' Send MSB first               
       REPEAT Bitcount                                                                
          outa[Dpin] := data >> (bitcount-1)                  ' Set output           
          data := data << 1                              ' Shift value right    
          !outa[Cpin]                                      ' cycle clock          
          !outa[Cpin]                                                             
          waitcnt(1000 + cnt)                              ' delay                
    outa[Dpin]~                                            ' Set data to low
     

  
PRI nm_core_CPU_carry: YesNo
  return carry

  
PRI nm_core_CPU_uninstall(bank)
  '/**
  ' * This function is called by CPU.removeVP. 
  ' * Remove a virtual peripheral from its slot. The VP will stop executing
  ' * immediately and its slot will be available to install a different VP.
  ' *
  ' * @param bank the rambank allocated to the virtual peripheral that must stop.
  ' */


PRI nm_core_CPU_install(routine)
  'Appears not to be used

  
PRI nm_core_CPU_nap(period)
  'If the Prop cannot support nap, then just do a delay
  '/**
  ' * Put the Prop into sleep mode for a period of time. The accuracy
  ' * of the sleep timer is very poor. See the manual for more information.
  ' *
  ' * @param period specifies the period for which the Prop should sleep. The
  ' * following values are valid:
  ' *
  ' * 0 - 16ms
  ' * 1 - 32ms
  ' * 2 - 64ms
  ' * 3 - 128ms
  ' * 4 - 256ms
  ' * 5 - 512ms
  ' * 6 - 1024ms
  ' * 7 - 2048ms
  ' */


PRI nm_core_CPU_setOutput(portPin)
  '/**
  ' * Make a pin an output.
  ' * The specified pin will be converted to an output. More than one pin can
  ' * be specified using the | operator, as long as all pins are on the same
  ' * port.
  ' *
  ' * @param portPin the pin(s) to make into an output.
  ' */
  if portPin < 256
    dira[7..0] |= portPin
  else
    portPin &= $FF
    dira[15..8] |= portPin
    

PRI nm_core_CPU_setInput(portPin)
  '/**
  ' * Make a pin an input.
  ' * The specified pin will be converted to an input. More than one pin can
  ' * be specified using the | operator, as long as all pins are on the same
  ' * port.
  ' *
  ' * @param portPin the pin(s) to make into an input.
  ' */
  if portPin < 256
    dira[7..0] &= (255 ^ portPin)
  else
    portPin &= $FF
    dira[15..8] &= (255 ^ portPin)


PRI nm_core_CPU_readSRAM(address): value
  '/**
  ' * Reads a byte from SRAM at the specified address.
  ' *
  ' * @param address the address to read the byte from.
  ' * @return the byte read from the SRAM.
  ' */
  return byte[address]
  

PRI nm_core_CPU_writeSRAM(address,value)
  '/**
  ' * Writes a byte into the SRAM at the specified address.
  ' *
  ' * @param address the address to write the byte to.
  ' * @param value value to write.
  ' */
  byte[address] := value
  

CON
  propSDA = 29
  propSCL = 28


PRI nm_core_EEPROM_eeRead(address): value | ack,temp
  '/**
  ' * Reads a byte from the EEPROM at the specified address.
  ' *
  ' * @param address the address to read the byte from.
  ' * @return the byte read from the EEPROM.
  ' */
  {{ adapted from object BS2_functions Pub Read(address, size) }}
  value~
  dira[propSCL]~~
  outa[propSCL]~~
  dira[propSDA]~
  outa[propSDA]~
  dira[propSDA]~~
  nm_core_CPU_shiftout(propSDA, propSCL, %10100000, SHIFT_MSB, 8)
  dira[propSDA]~
  ack := nm_core_CPU_shiftin(propSDA, propSCL, PRE_CLOCK_LSB, 1)
  nm_core_CPU_shiftout(propSDA, propSCL, address >> 8, SHIFT_MSB, 8)
  dira[propSDA]~
  ack := nm_core_CPU_shiftin(propSDA, propSCL, PRE_CLOCK_LSB, 1)
  nm_core_CPU_shiftout(propSDA, propSCL, address, SHIFT_MSB, 8)
  dira[propSDA]~
  ack := nm_core_CPU_shiftin(propSDA, propSCL, PRE_CLOCK_LSB, 1)
  dira[propSCL]~~
  outa[propSCL]~~
  dira[propSDA]~
  outa[propSDA]~
  dira[propSDA]~~
  nm_core_CPU_shiftout(propSDA, propSCL, %10100001, SHIFT_MSB, 8)
  dira[propSDA]~
  ack := nm_core_CPU_shiftin(propSDA, propSCL, PRE_CLOCK_LSB, 1)
  dira[propSDA]~
  value := nm_core_CPU_shiftin(propSDA, propSCL, PRE_CLOCK_MSB, 8)
  ack := nm_core_CPU_shiftin(propSDA, propSCL, PRE_CLOCK_LSB, 1)
  outa[propSDA]~
  dira[propSDA]~~
  outa[propSCL]~~
  dira[propSDA]~
        

PRI nm_core_EEPROM_eeWrite(address,value) | ack 
  '/**
  ' * Writes a byte into the EEPROM at the specified address.
  ' *
  ' * @param address the address to write the byte to.
  ' * @param value value to write.
  ' */
  {{ adapted from object BS2_functions Pub Write(address, value, size) }}
  outa[propSCL]~~
  dira[propSCL]~~
  dira[propSDA]~
  outa[propSDA]~
  dira[propSDA]~~
  nm_core_CPU_shiftout(propSDA, propSCL, %10100000, SHIFT_MSB, 8)
  dira[propSDA]~
  ack := nm_core_CPU_shiftin(propSDA, propSCL,OnClock,1)
  nm_core_CPU_shiftout(propSDA, propSCL, address >> 8, SHIFT_MSB, 8)
  dira[propSDA]~
  ack := nm_core_CPU_shiftin(propSDA, propSCL,OnClock, 1)
  nm_core_CPU_shiftout(propSDA, propSCL, address, SHIFT_MSB, 8)
  dira[propSDA]~
  ack := nm_core_CPU_shiftin(propSDA, propSCL,OnClock,1)
  nm_core_CPU_shiftout(propSDA, propSCL, value, SHIFT_MSB, 8)
  dira[propSDA]~
  ack := nm_core_CPU_shiftin(propSDA, propSCL,OnClock, 1)
  outa[propSDA]~
  dira[propSDA]~~
  outa[propSCL]~~
  dira[propSDA]~
  waitcnt(clkfreq/200 + cnt)
           


CON
  'Uart registers
  nmUartDataH    = $02
  nmUartDataL    = $03
  nmUartInvert   = $05 '// 0 for normal, 1 for invert
  nmUartBaud     = $06 '// 1000000/(bps * 4.34)
  nmUartStopBits = $07 '// Tx only: 00000001 for 1 stop bit, 00000011 for 2 etc
  nmUartPort     = $0C '// 0 for Port0, 1 for Port1
  nmUartPin      = $0D '// bit mask (1 of 8 set)
  nmUartHSPort   = $0E '// 0 for Port0, 1 for Port1
  nmUartHSPin    = $0F '// bitmask (1 of 8 set)
  nmUartBufferH  = $0A
  nmUartBufferL  = $0B
  nmUartHead     = $08 '//Index of the next free byte in the buffer. 
  nmUartTail     = $09 '//Index of the oldest byte in the buffer.
                       '//If tailPointer == headPointer the buffer is empty.

  'Uart baud constants
  speed600   = 192  'The speed value for 600 baud.
  speed1200  = 96   'The speed value for 1200 baud.
  speed2400  = 48   'The speed value for 2400 baud.
  speed4800  = 24   'The speed value for 4800 baud.
  speed7200  = 16   'The speed value for 7200 baud.
  speed9600  = 12   'The speed value for 9600 baud.
  speed14400 = 8    'The speed value for 14400 baud.
  speed19200 = 6    'The speed value for 19200 baud.
  speed28800 = 4    'The speed value for 28800 baud.
  speed38400 = 3    'The speed value for 38400 baud.
  speed57600 = 2    'The speed value for 57600 baud.


PRI nm_core_Uart_txInit(bank)
  '/**
  ' * Initialize a Uart transmitter as background task. Its parameters are in vpRambank.
  ' * Called by Uart.start()
  ' *
  ' * @param bank The allocated area for the Uart registers.
  ' */
  bank -= nmPeriph0Bank
  'Access the proper register using vpRambank[bank+register]
  

PRI nm_core_Uart_rxInit(bank)
  '/**
  ' * Initialize a Uart receiver as background task. Its parameters are in vpRambank.
  ' * Called by Uart.start()
  ' *
  ' * @param bank The allocated area for the Uart registers.
  ' */
  bank -= nmPeriph0Bank
  'Access the proper register using vpRambank[bank+register]


PRI nm_core_Uart_rxRead(bank)
  '/**
  ' * Inform Uart receiver a byte has been grabbed from the buffer, so the reciever can do the handshaking.
  ' * Called by Uart.readByte()
  ' *
  ' * @param bank The allocated area for the Uart registers.
  ' */
  bank -= nmPeriph0Bank
  'Access the proper register using vpRambank[bank+register]


PRI nm_core_Callback_returnTwo
  'This appears not to be used

  
CON
  'Timer registers - Note that the Prop already has a running 32bits timer
  nmPeriphISRVec = $00
  nmLatchTimer1  = $06
  nmLatchTimer2  = $07
  nmLatchTimer3  = $08
  nmLatchTimer4  = $09
  nmTimer1       = $02
  nmTimer2       = $03
  nmTimer3       = $04
  nmTimer4       = $05


PRI nm_core_Timer_init(bank)
  '/**
  ' * Initialize the 32bit counter. The 32bit nmTimer register should increment every 8.68 usec.
  ' * Called by Timer.start()
  ' *
  ' * @param bank The allocated area for the Timer registers.
  ' */


PRI nm_core_Timer_latch(bank)
  '/**
  ' * Latch the 32bit counter in the latch registers.
  ' * Called by Timer.tickLo()
  ' *
  ' * @param bank The allocated area for the Timer registers.
  ' */
  bank -= nmPeriph0Bank
  'Access the proper register using vpRambank[bank+register]
  'This probably needs a lock
  vpRambank[bank + nmLatchTimer1] := vpRambank[bank + nmTimer1]
  vpRambank[bank + nmLatchTimer2] := vpRambank[bank + nmTimer2]
  vpRambank[bank + nmLatchTimer3] := vpRambank[bank + nmTimer3]
  vpRambank[bank + nmLatchTimer4] := vpRambank[bank + nmTimer4]
  

CON
  'PWM registers
  nmPWMPort = $0E '// 0 for Port0, 1 for Port1
  nmPWMPin  = $0F '// bitmask (1 of 8 set)
  'not defined in java sources but we can use unused registers
  nmPWMhighTime = $0A
  nmPWMlowTime  = $0B
  nmPWMhighZero = $0C
  nmPWMlowZero  = $0D


PRI nm_core_PWM_install(bank) | pin,low,high,duty,scale
  '/**
  ' * Initialize a PWM as background task. Its parameters are in vpRambank.
  ' * Called by PWM.start()
  ' *
  ' * @param bank The allocated area for the PWM registers.
  ' */
  bank -= nmPeriph0Bank
  'Access the proper register using vpRambank[bank+register]
  {{ adapted from object BS2_functions Pub PWM_Set(Pin, Duty, Resolution) | Scale }}
  {{ Produces constant PWM on the specified pin }}
  {{ Needs to calculate scale }}
  if vpRambank[bank + nmPWMPort] == 0
    pin := >|vpRambank[bank + nmPWMPin]
  else
    pin := 8 + >|vpRambank[bank + nmPWMPin]
  low := vpRambank[bank + nmPWMlowTime]
  high := vpRambank[bank + nmPWMhighTime]
  duty := (high*100) / (low + high)
  if duty == 0                                         ' freq = 0 turns off square wave
     ctra := 0                                          ' Set CTRA/B to 0
     dira[pin]~                                         ' Make pin input
  else
      'scale := 2_147_483_647 / (1<< (Resolution-1))  ' Calculate scale
      ctra[30..26] := %00110      ' Set ctra to DUTY mode
      ctra[5..0] := pin           ' Set ctra's APIN
      frqa := duty * scale        ' Set frqa register
      dira[pin]~~                 ' set direction


PRI nm_core_PWM_updateInternal(bank,highTime,lowTime,highZero,lowZero)
  '/**
  ' * Update the PWM duty cycle. Its parameters are in vpRambank. 
  ' * Called by PWM.update()
  ' *
  ' * @param bank The allocated area for the PWM registers.
  ' * @param highTime the number of 8.68us the waveform should be high.
  ' * @param lowTime the number of 8.68us the waveform should be low.
  ' * @param highZero Boolean value (always true) that tells wether highTime != 0.
  ' * @param lowZero Boolean value (always true) that tells wether lowTime != 0.
  ' */
  bank -= nmPeriph0Bank
  'Access the proper register using vpRambank[bank+register]
  vpRambank[bank + nmPWMlowTime] := lowTime
  vpRambank[bank + nmPWMhighTime] := highTime
  vpRambank[bank + nmPWMlowZero] := lowZero
  vpRambank[bank + nmPWMhighZero] := highZero
  {{ Need to update ctra and frqa }}
  

CON
  'DAC registers
  nmDacValue = $0B '// Current ADC value.
  nmDacPort  = $0E '// 0 for Port0, 1 for Port1
  nmDacPin   = $0F '// bitmask (1 of 8 set)


PRI nm_core_DAC_install(bank)
  '/**
  ' * Initialize a DAC as background task. Its parameters are in vpRambank.
  ' * Called by DAC.start()
  ' *
  ' * @param bank The allocated area for the DAC registers.
  ' */
  bank -= nmPeriph0Bank
  'Access the proper register using vpRambank[bank+register]
  {{ Need some 8bit DAC routine }}


CON
  'ADC registers
  nmAdcValue = $0B '// Current ADC value.
  nmInPort   = $0C '// 0 for Port0, 1 for Port1
  nmInPin    = $0D '// bitmask (1 of 8 set)
  nmOutPort  = $0E '// 0 for Port0, 1 for Port1
  nmOutPin   = $0F '// bitmask (1 of 8 set)


PRI nm_core_ADC_install(bank)
  '/**
  ' * Initialize an ADC as background task. Its parameters are in vpRambank.
  ' * Called by ADC.start()
  ' *
  ' * @param bank The allocated area for the ADC registers.
  ' */
  bank -= nmPeriph0Bank
  'Access the proper register using vpRambank[bank+register]
  {{ Need some 8bit ADC routine }}

  
PRI nm_core_Terminal_getByte
  '/**
  ' * Request a byte from the debug terminal.
  ' * Called by Terminal.internalGetChar()
  ' *
  ' * @return The byte sent by the debug terminal.
  ' */
  return tv.key
  

PRI nm_page_
  'Since the Prop has no pages, simply ignore ???
 
{{
 *
 */
}}

PUB jvmStep 
{{
/**
 * This is a debug control command that allows JVM to run until a breakpoint is hit.
 * @returns zero if break happened normally.
 */
__ int jvmStep();
}}
  result := jvmSpinDebug(0)

PUB jvmSingleStepBytecode 
{{
/**
 * This is a debug control command that allows JVM to run one bytecode "instruction" (opcode/operand).
 * An instruction is of the form OPCODE+OPERANDS, i.e. "bipush 6" or bytecode sequence $08,$06.
 * Many instructions take information from the stack and/or manipulate stack contents.
 * The state of the stack must be reasonable before many instructions can be executed.
 * The method blocks until the instruction is complete or an error occured. 
 */
__ int jvmSingleStepBytecode();
}}
  result := jvmSpinDebug(1)

PUB jvmRun 
{{
/**
 * This is a debug control command that allows JVM to run the bytecode program without debug control. 
 */
__ void jvmRun();
}}
  debugMsDelay := 0
  JvmInit    
  JvmSpinLoop

PUB jvmStop 
{{
/**
 * This is a debug control command. 
 * Stop the JVM from executing the next instruction and abort the current one if possible.
 * Hmm. This wont work unless jvm is running in a COG ..... 
 */
__ void jvmStop();
}}
  gAbort := true
  
PUB clearAllJvmBreakpoints 
{{
/**
 * This is a debug control command that clears all breakpoints.
 * The breakpoint queue pointer will be 0 upon return.
 */
__ void clearAllJvmBreakpoints();
}}
  gBreakpointHead := 0
  repeat BREAKPOINT_SIZE
    gBreakpointArray[gBreakpointHead++] := 0
    gBreakpointHead //= BREAKPOINT_SIZE

PUB deleteJvmBreakpoint(bppc) | ii, jj 
{{
/**
 * This is a debug control command that lets the user delete a breakpoint.
 * @param bppc is the program counter byte array instruction address.
 * @returns zero if remove breakpoint failed
 */
__ int deleteJvmBreakpoint(bppc);
}}
  result := 0

  ii := -1  
  repeat while(++ii < BREAKPOINT_SIZE)
    if gBreakpointArray[ii] == bppc
      ' shuffle breakpoints
      repeat while(++ii < BREAKPOINT_SIZE)
        gBreakpointArray[ii-1] := gBreakpointArray[ii]
      tv.str(string($d,"deleted bp "))
      tv.hex(bppc, 4)
      return 1


PUB setJvmBreakpoint(bppc) 
{{
/**
 * This is a debug control command that lets the user set one breakpoint.
 * The breakpoint list is an array of BREAKPOINT_SIZE words that behaves like a queue.
 * Only BREAKPOINT_SIZE entries are allowed at one time. The newest entry will replace
 * the oldest entry.
 * @param pc is the program counter byte array instruction address.
 * @returns position of last breakpoint in queue  
 */
__ int setJvmBreakpoint(bppc);
}}
  tv.Str(String($d,"line number $"))
  tv.hex(bppc,4)
  tv.out(" ")
  result := gBreakpointArray[gBreakpointHead] := bppc
  'tv.hex(result,4)
  'tv.out(" ")
  gBreakpointHead++
  'result := gBreakpointHead
  'tv.hex(result,4)
  'tv.out(" ")
  result := gBreakpointHead //= BREAKPOINT_SIZE
  tv.hex(result,4)
                
PUB getJvmBreakpoint(num) 
{{
/**
 * This is a debug control command that lets the user get the pc of a breakpoint.
 * If the breakpoint input parameter is 0, get the last breakpoint, else get breakpoint specified.
 * The breakpoint number should be <= maximum allowed breakpoints.
 * If the number is greater than maximum, the breakpoint will be the mod of of the number.
 * @param num is the 1-base breakpoint number to get. If num is 0 get the last breakpoint set.
 * @returns pc of selected breakpoint or zero if breakpoint not found. 
 */
__ int getJvmBreakpoints(num);
}}
  if num > 0
    num--
    num //= BREAKPOINT_SIZE
    result := gBreakpointArray[num]

PUB getMaxJvmBreakpoints 
{{
/**
 * This is a debug control command that tells user maximum number of breakpoints allowed.
 * @returns maximum number of breakpoints. 
 */
__ int getMaxJvmBreakpoints();
}}
  result := BREAKPOINT_SIZE

PUB isBreakPoint(bppc) | ii
{{
/**
 * find out if PC is on a breakpoint
 * @param pc the program counter value to check for breakpoint
 * @returns non-zero or true if pc is a breakpoint
 */
__ bool isBreakPoint(pc);
}}
  result := false
  if gGetOverBreak
    gGetOverBreak := 0
    return false
  repeat ii from 0 to BREAKPOINT_SIZE-1
    if gBreakpointArray[ii] == bppc
      return true

PUB listJvmBreakpoints | val, ii
{{
/**
 * list all breakpoints
 */
__ void listJvmBreakpoints();
}}
  result := 0
  repeat ii from 1 to BREAKPOINT_SIZE
    val := getJvmBreakpoint(ii)
    if val <> 0
      tv.str(string($d,"Breakpoint "))
      tv.dec(ii)
      tv.str(string(" @PC "))
      tv.hex(val, 4)
      tv.str(string(" JemPC "))
      tv.hex(val-bytecodeStart, 4)
      result++

  if result == 0
    tv.str(string($d,"No Breakpoints Set"))
        

PUB getJvmAbortPointer 
{{
/**
 * This is command lets caller get the abort pointer.
 * This variable would be used to cause an execution abort for debugging.
 * @returns address of the JVM engine's abort variable.
 */
__ int getJvmAbortPointer();
}}
  result := @gAbort
  
  
PUB getJvmPC 
{{
/**
 * This is a debug status command that returns the value of the bytecode offset program counter. 
 * This value will be the address in memory of the current PC instruction. 
 * @returns value of the program counter. 
 */
__ int getJvmPC();
}}
  result := pc
  
PUB getJvmSP 
{{
/**
 * This is a debug status command that returns the value of the stack pointer. 
 * This value will be the address in memory of the current SP location. 
 * @returns value of the stack pointer. 
 */
__ int getJvmSP();
}}
  result := sp

PUB getJvmSB 
{{
/**
 * This is a debug status command that returns the value of the stack base pointer. 
 * This value will be the address in memory of the current SB location. 
 * This value is a representation of the JVM's internal BP or base pointer. 
 * @returns value of the stack base pointer. 
 */
__ int getJvmSB();
}}
  result := bp
  
PUB getJvmFP 
{{
/**
 * This is a debug status command that returns the value of the Frame Pointer. 
 * The FP is defined as the beginning of the stack for the currently executing method. 
 * @returns value of the frame pointer. 
 */
__ int getJvmFP();
}}
  result := fp
  
PUB getJvmMP 
{{
/**
 * This is a debug status command that returns the relative bytecode offset value of the currently running 
 * Method's start address. This value is is the constant table method start address + the bytecode base address. 
 * @returns value of the method pointer. 
 */
__ int getJvmMP();
}}
  result := mb

PUB getJvmHP 
{{
/**
 * This is a debug status command that returns the value of the heap start pointer. 
 * @returns value of the heap start pointer. 
 */
__ int getJvmHP();
}}
  result := getJvmHeapStart
  
PUB getJvmHeapStart 
{{
/**
 * This is a debug status command that returns the current start position of the Heap pointer. 
 * @returns value of the heap pointer lowest memory position. 
 */
__ int getJvmHeapStart();
}}
  result := heap
  
PUB getJvmHeapEnd 
{{
/**
 * This is a debug status command that returns the current end position of the Heap pointer. 
 * @returns value of the heap pointer highest memory location.
 */
__ int getJvmHeapEnd();
}}
  result := HEAP_BASE

       